# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

from __future__ import unicode_literals
import logging
import logging.handlers

from traits.api import Any, Int, Str, Unicode
from colorlog import ColoredFormatter
import ansiconv

from automate.service import AbstractUserService

__all__ = ['LogStoreService']


class LogStoreService(AbstractUserService):

    """
        Provides interface to log output. Used by WebService.
    """
    autoload = True

    #: Log level
    log_level = Int(logging.INFO)

    #: Log length
    log_length = Int(100)

    #: The most recent log line is always updated here.
    #:t Subscription to this attribute can be used to follow new log entries.
    most_recent_line = Unicode

    format = Str('%(log_color)s%(asctime)s %(name)s %(message)s %(reset)s')
    html_format = Str('%(white)s%(asctime)s %(log_color)s%(name)s%(white)s %(message)s %(reset)s')
    time_format = Str('%H:%M:%S')

    _loghandler = Any

    def _log_level_changed(self, new):
        self._loghandler.setLevel(new)

    @staticmethod
    def html_fix(s):
        return s.replace('<', '&lt;').replace('>', '&gt;')


    def setup(service):
        html_formatter = ColoredFormatter(service.html_format, datefmt='%H:%M:%S')

        class MyBufferingHandler(logging.handlers.BufferingHandler):

            def flush(self):
                del self.buffer[:int(.25 * self.capacity)]

            def emit(self, record):
                super(MyBufferingHandler, self).emit(record)
                service.most_recent_line = ansiconv.to_html(service.html_fix(html_formatter.format(record))) + '\n'

        service._loghandler = loghandler = MyBufferingHandler(service.log_length)
        loghandler.setLevel(service.log_level)
        loghandler.setFormatter(ColoredFormatter(service.format, datefmt=service.time_format))
        service.system.logger.addHandler(loghandler)

    def lastlog(self, lines=20, format='', html=True):
        handler = self._loghandler
        if format:
            formatter = ColoredFormatter(format, datefmt=self.time_format)
        else:
            formatter = ColoredFormatter(self.html_format if html else self.format, datefmt=self.time_format)

        rv = u'\n'.join([formatter.format(i) for i in handler.buffer[-lines:]])

        if html:
            rv = ansiconv.to_html(self.html_fix(rv))
        else:
            rv = ansiconv.to_plain(rv)
        return rv

    def cleanup(self):
        return



# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of automate-rpio.
#
# automate-rpio is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# automate-rpio is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with automate-rpio.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals
from traits.api import CInt, CBool, Instance, Any, CFloat, Property
from automate.service import AbstractSystemService
from automate.statusobject import AbstractActuator


class RpioActuator(AbstractActuator):

    """
        Boolean-valued actuator for setting Raspberry Pi GPIO port statuses (on/off).
    """

    _status = CBool(transient=True)

    #: GPIO port id
    port = CInt

    #: Set to True to have inversed status value
    inverted = CBool(False)

    _hw_service = Instance(AbstractSystemService, transient=True)

    view = AbstractActuator.view + ["port"]

    def setup(self):
        self._hw_service = self.system.request_service('RpioService')
        self._hw_service.enable_output_port(self.port)

    def _port_changed(self, old, new):
        if not self._hw_service:
            return
        if old:
            self._hw_service.disable_output_port(old)

        self._hw_service.enable_output_port(new)

    def _status_changed(self):
        self.logger.debug("%s gpio actuator status changed %s", self.name, repr(self._status))
        self._hw_service.set_output_port_status(self.port, self._status if not self.inverted else not self._status)


class RpioPWMActuator(AbstractActuator):

    """
        Actuator to control PWM (pulse-width-modulation) ports on Raspberry pi GPIO.

        Status range 0...1

        This is not recommended to be used because RPIO PWM implementation is not very well behaving.
        I recommend to use ArduinoPWMActuator with an Arduino loaded with StandardFirmata. It's much more
        stable and robust solution.
    """

    _status = CFloat(transient=True)

    #: GPIO port number
    port = CInt

    #: RPIO PWM DMA channel
    dma_channel = CInt(0)

    #: PWM frequency (Hz)
    frequency = CFloat(50.)

    _subcycle = Property(trait=CInt, transient=True)

    view = AbstractActuator.view + ["port"]

    _pwm = Any(transient=True)

    _hw_service = Instance(AbstractSystemService, transient=True)

    def _get__subcycle(self):
        return int(1. / self.frequency * 1e6)

    def setup(self):
        self._hw_service = self.system.request_service('RpioService')
        self._pwm = self._hw_service.get_pwm_module()
        self._port_changed(None, self.port)

    def _port_changed(self, old, new):
        if not self._hw_service:
            return

        if old:
            self._pwm.clear_channel_gpio(self.dma_channel, old)

        self._pwm.init_channel(self.dma_channel, self._subcycle)
        new = min(0.999, max(0., self._status))
        self._pwm.add_channel_pulse(self.dma_channel, self.port, 0, int(round(new * self._subcycle / 10.)))

    def _frequency_changed(self):
        self._port_changed(self.port, self.port)

    def _status_changed(self):
        self.logger.debug("PWMActuator %s change %s", self, self._status)

        PWM = self._pwm

        PWM.clear_channel_gpio(self.dma_channel, self.port)
        new = min(0.999, max(0., self._status))
        PWM.add_channel_pulse(self.dma_channel, self.port, 0, int(round(new * self._subcycle / 10.)))



# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of automate-webui.
#
# automate-webui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# automate-webui is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with automate-webui.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

from __future__ import division
from __future__ import unicode_literals
import re
import keyword

import django.forms as forms
from django.core.urlresolvers import reverse
from django.contrib import messages

from traits.api import TraitError

from automate.callable import AbstractCallable
from automate.program import Program

fieldtypes = {int: forms.IntegerField, float: forms.FloatField, bool: forms.BooleanField}


class TextForm(forms.Form):
    """
    Larger text edit form shown in info panel for user_editable text objects
    """
    name = forms.CharField(widget=forms.HiddenInput())
    status = forms.CharField(widget=forms.Textarea(attrs={'rows': 2, 'class': 'mytextbox'}),
                             required=False)

    def __init__(self, data=None, source='', **kwargs):
        super(TextForm, self).__init__(data, **kwargs)
        from crispy_forms.helper import FormHelper
        from crispy_forms.layout import Layout, Submit
        self.helper = FormHelper()
        self.helper.layout = Layout('name', 'status',
                                    Submit('set_status', 'Set', css_class='textform_submit'))
        self.helper.form_class = "form-horizontal"
        self.helper.form_action = reverse('set_status') + '?source=%s' % source


class QuickEdit(forms.Form):
    """
    Small quick edit forms that are used in main views to edit object statuses
    """
    name = forms.CharField(widget=forms.HiddenInput())
    status = forms.BooleanField()

    def __init__(self, data=None, source='', sensor=None, **kwargs):
        super(QuickEdit, self).__init__(data, **kwargs)
        self.fields['status'].label = ''
        from crispy_forms.helper import FormHelper
        from crispy_forms.bootstrap import FieldWithButtons
        from crispy_forms.layout import Layout, Submit
        self.helper = FormHelper()
        self.helper.layout = Layout('name', FieldWithButtons('status', Submit('set_status', '>')))
        self.helper.form_class = "form-inline quickedit"
        self.helper.form_action = reverse('set_status') + '?source=%s' % source


class NumericQuickEdit(QuickEdit):
    status = forms.Field()

    def __init__(self, data=None, source='', sensor=None, **kwargs):
        from crispy_forms.bootstrap import FieldWithButtons
        from crispy_forms.layout import Layout, Submit, Field
        super(NumericQuickEdit, self).__init__(data, source, sensor, **kwargs)
        if sensor.is_finite_range:
            field = self.fields['status']
            field.sensor_name = sensor.name
            field.min_value = self.num_type(sensor.value_min)
            field.max_value = self.num_type(sensor.value_max)
            if self.num_type is int:
                field.step = 1
            else:
                field.step = (sensor.value_max - sensor.value_min) / 1000.

            self.helper.layout = Layout('name', Field('status', template='slider.html'))
        else:
            self.helper.layout = Layout('name',
                                        FieldWithButtons('status', Submit('set_status', '>')))


class IntQuickEdit(NumericQuickEdit):
    num_type = int
    status = forms.IntegerField()


class FloatQuickEdit(NumericQuickEdit):
    num_type = float
    status = forms.FloatField()


class StrQuickEdit(QuickEdit):
    status = forms.CharField(required=False)


class BoolQuickEdit(QuickEdit):
    status = forms.BooleanField()


QUICK_EDITS = dict(str=StrQuickEdit, unicode=StrQuickEdit, int=IntQuickEdit, float=FloatQuickEdit,
                   bool=BoolQuickEdit)


class BaseForm(forms.Form):
    """
    Base crispy form for full-page forms such as LoginForm and SystemObjectForms
    """

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop('request', None)
        super(BaseForm, self).__init__(*args, **kwargs)
        from crispy_forms.helper import FormHelper
        from crispy_forms.layout import Submit
        self.helper = FormHelper()
        self.helper.add_input(Submit('submit', 'Submit'))
        self.helper.add_input(Submit('cancel', 'Cancel'))
        self.helper.form_class = 'form-horizontal'


class LoginForm(BaseForm):
    username = forms.CharField()
    password = forms.CharField(widget=forms.PasswordInput())

    def clean(self):
        cleaned_data = super(LoginForm, self).clean()
        if (cleaned_data.get('username') != self.auth[0]
            or cleaned_data.get('password') != self.auth[1]):
            raise forms.ValidationError('Username or password is not correct', code='invalid')
        return cleaned_data


class CmdForm(forms.Form):
    cmd = forms.CharField(label='Command', widget=forms.TextInput(attrs={'autocomplete': 'off'}))

    def __init__(self, *args, **kwargs):
        textarea = kwargs.pop('textarea', False)
        super(CmdForm, self).__init__(*args, **kwargs)
        from crispy_forms.helper import FormHelper
        self.helper = FormHelper()
        from crispy_forms.bootstrap import FieldWithButtons
        from crispy_forms.layout import Layout, Submit
        if textarea:
            self.fields['cmd'].widget = forms.Textarea(attrs=dict(cols=30, rows=3))
            self.helper.layout = Layout('cmd', Submit('run_cmd', 'Run'))
        else:
            self.helper.layout = Layout(FieldWithButtons('cmd', Submit('run_cmd', 'Run')))
        self.helper.form_class = "form-horizontal"


class SystemObjectForm(BaseForm):
    name = forms.CharField()
    description = forms.CharField(widget=forms.Textarea(attrs=dict(rows=3)), required=False)
    tags = forms.MultipleChoiceField(required=False)
    new_tags = forms.CharField(required=False, initial='')

    def new(self, name):
        raise NotImplementedError

    def get_object(self, name):
        obj = self.system.namespace.get(name, None)
        self.instance = obj or self.new(self.cleaned_data['name'])
        return self.instance

    def save(self, objname):
        obj = self.get_object(objname)
        for key, value in self.cleaned_data.items():
            if key in self.changed_data:
                if key == 'new_tags':
                    for tag in value.split(','):
                        obj.tags.add(tag)
                    continue
                try:
                    setattr(obj, key, value)
                except Exception as e:
                    messages.error(self.request,
                                   'Error was caught when trying to set %s: %s' % (key, e))

    def clean_name(self):
        name = self.cleaned_data['name']
        if 'name' in self.changed_data and name in self.system.namespace:
            raise forms.ValidationError('Name already in %s namespace' % self.system.name)
        if not re.match("^[_A-Za-z][_a-zA-Z0-9]*$", name):
            raise forms.ValidationError('Illegal characters')
        if keyword.iskeyword(name):
            raise forms.ValidationError('Name is python keyword')
        return name

    def setup_system(self, system):
        self.system = system
        self.load_from_system()

    def load_from_system(self):
        all_tags = list(self.system.all_tags)
        self.fields['tags'].choices = [(i, i) for i in all_tags if i]

    @property
    def class_name(self):
        if hasattr(self, 'instance'):
            return self.instance.class_name
        else:
            return 'New ' + self.__class__.__name__.replace('Form', '')

    def load(self, objname):
        if objname:
            # fill those defined fields that are without initial value
            self.obj = obj = self.system.namespace[objname]
            self.instance = obj
            self.fields['tags'].initial = list(obj.tags)
            for key, value in self.fields.items():
                if value.initial is None:
                    value.initial = getattr(obj, key)


class ProgramForm(SystemObjectForm):
    priority = forms.FloatField()

    def load(self, programname):
        if programname:
            prog = self.system.namespace[programname]

            for key in prog.callables + ['triggers', 'exclude_triggers']:
                cll = getattr(prog, key)
                if isinstance(cll, AbstractCallable):
                    val = cll.give_str_indented()
                elif isinstance(cll, set):
                    val = repr(set(cll))

                self.fields[key] = forms.CharField(
                    initial=val, required=False,
                    widget=forms.Textarea(attrs=dict(rows=val.count('\n') + 1)))
        super(ProgramForm, self).load(programname)

    def clean(self):
        from automate.common import LogicStr

        obj = getattr(self, 'obj', None)
        if obj:
            for key in obj.callables + ['triggers', 'exclude_triggers']:
                if key in self.changed_data:
                    try:
                        LogicStr().validate(list(self.system.programs)[0], None,
                                            self.cleaned_data[key])
                    except TraitError:
                        self.add_error(key, 'Invalid value')

        return super(ProgramForm, self).clean()

    def new(self, progname):
        prog = Program(self.cleaned_data.pop('name'))
        prog.setup_system(self.system)
        return prog

    def save(self, progname=None):
        prog = self.get_object(progname)

        for key in prog.callables + ['triggers', 'exlude_triggers']:
            if key in self.changed_data:
                setattr(prog, key + '_str', self.cleaned_data.pop(key))

        super(ProgramForm, self).save(prog.name)


class StatusObjectForm(ProgramForm):
    safety_delay = forms.FloatField()
    safety_mode = forms.ChoiceField(choices=[(i, i) for i in ['rising', 'falling', 'both']])
    change_delay = forms.FloatField()
    change_mode = forms.ChoiceField(choices=[(i, i) for i in ['rising', 'falling', 'both']])

    def load(self, objname=None):
        if objname is None:
            # Object type field for new objects
            self.fields['object_type'] = forms.ChoiceField(choices=[(i, i) for i in self.types])
        else:
            obj = self.get_object(objname)
            for attr_name in obj.view:
                if attr_name not in self.fields:
                    init_value = getattr(obj, attr_name)
                    value_type = type(init_value)
                    self.fields[attr_name] = fieldtypes.get(
                        value_type, forms.CharField)(initial=init_value, required=False)
        super(StatusObjectForm, self).load(objname)

    def new(self, name):
        type = self.cleaned_data.pop('object_type')
        Type = getattr(self.module, type)
        obj = Type(self.cleaned_data.pop('name'))
        obj.setup_system(self.system)
        return obj

    @property
    def abstract_class(self):
        raise NotImplementedError

    @property
    def module(self):
        raise NotImplementedError

    @property
    def types(self):
        return sorted(
            [key for key, value in self.module.__dict__.items()
             if isinstance(value, type) and issubclass(value, self.abstract_class)
             and not key.startswith('Abstract')])


class SensorForm(StatusObjectForm):
    @property
    def abstract_class(self):
        from automate.statusobject import AbstractSensor
        return AbstractSensor

    @property
    def module(self):
        import automate.sensors
        return automate.sensors


class ActuatorForm(StatusObjectForm):
    @property
    def abstract_class(self):
        from automate.statusobject import AbstractActuator
        return AbstractActuator

    @property
    def module(self):
        import automate.actuators
        return automate.actuators


FORMTYPES = dict(program=ProgramForm, sensor=SensorForm, actuator=ActuatorForm)



# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

from __future__ import unicode_literals, division

import datetime
from builtins import bytes

from future import standard_library

standard_library.install_aliases()

import re
import threading
import xmlrpc.client
import socket
import subprocess

from http.client import HTTPException

from traits.api import (CList, Any, Property, Set, Bool, Event, CBool, on_trait_change, cached_property)

from automate.callable import AbstractCallable
from automate.common import deep_iterate, get_modules_all
from automate.statusobject import StatusObject
from automate.common import (threaded, thread_start, is_iterable)


class Empty(AbstractCallable):

    """
        Do nothing but return None. Default action in Programs.

        Usage::

            Empty()
    """
    test = Bool(False)

    def call(self, caller, **kwargs):
        return None


class AbstractAction(AbstractCallable):

    """
        Abstract base class for actions (i.e. callables that do something but
        do not necessarily return anything.
    """
    status = Bool(True)


class Attrib(AbstractAction):

    """
        Give specified attribute of a object.

        :param no_eval bool: if True, evaluation of object is skipped -- use this to access attributes of SystemObjects

        Usage & example::

            Attrib(obj, 'attributename')
            Attrib(sensor_name, 'status', no_eval=True)


    """

    @property
    def method(self):
        return self._args[1]

    def call(self, caller, **kwargs):
        no_eval = self._kwargs.get('no_eval', False)
        if not caller:
            return
        if no_eval:
            obj = self.obj
        else:
            obj = self.call_eval(self.obj, caller, **kwargs)
        attr = self.call_eval(self.method, caller, **kwargs)
        return getattr(obj, attr, None)


class Method(AbstractAction):

    """
        Call method in an object with specified args

        Usage::

            Method(obj, 'methodname')
    """

    @property
    def method(self):
        return self._args[1]

    @property
    def args(self):
        try:
            return self._args[2:]
        except IndexError:
            return ()

    def call(self, caller, **kwargs):
        if not caller:
            return
        kwargs = self._kwargs.copy()
        if not kwargs.pop('no_caller', False):
            arglist = [caller] + list(self.args)
        else:
            arglist = self.args
        return getattr(self.call_eval(self.obj, caller, **kwargs), self.call_eval(self.method, caller, **kwargs))(*arglist, **kwargs)


class Func(AbstractAction):

    """
        Call function with given arguments.

        Usage & example::

            Func(function, *args, **kwargs)
            Func(time.sleep, 2)

        :param bool add_caller: if True, then caller program is passed as first argument.
    """

    @property
    def args(self):
        return tuple(self._args[1:])

    def call(self, caller, **kwargs):
        if not caller:
            return
        _kwargs = {k: self.call_eval(v, caller, **kwargs) for k, v in list(self._kwargs.items())}
        return_value = _kwargs.pop('return_value', True)
        args = [self.call_eval(i, caller, return_value=return_value, **kwargs) for i in self.args]
        if _kwargs.pop('add_caller', False):
            arglist = [caller] + list(args)
        else:
            arglist = args
        self.logger.debug("Func %s %s %s", self.obj, arglist, _kwargs)
        try:
            return self.call_eval(self.obj, caller, **kwargs)(*arglist, **_kwargs)
        except Exception as e:
            self.logger.error('Exception occurred in %s: %s', self, e)


class OnlyTriggers(AbstractCallable):

    """
    Baseclass for actions that do not have any targets (i.e. almost all actions).
    """

    def _give_targets(self):
        return None


class Log(OnlyTriggers):

    """
        Print callable argument outputs / other arguments to the log.

        Usage::

            Log(object1, object2, 'string1'...)

        :param str log_level: Log level (i.e. logging function name) (default 'info')
    """

    default_log_level = 'info'

    def call(self, caller, **kwargs):
        log_level = self._kwargs.get('log_level', self.default_log_level)
        l = []
        for o in self.objects:
            l.append(self.call_eval(o, caller, **kwargs))
        getattr(self.logger, log_level)(*l)
        return True


class Debug(Log):

    """
        Same as :class:`.Log` but with debug logging level.
    """
    default_log_level = 'debug'


class ToStr(OnlyTriggers):

    """
        Return string representation of given arguments evaluated.
        Usage::

            ToStr('formatstring {} {}', callable1, statusobject1)

        :param bool no_sub: if True, removes format string from argument list. Then usage is simply:

        .. code-block:: python

            ToStr(callable1, statusobject1, no_sub=True)
    """

    def call(self, caller, **kwargs):
        _kwargs = self._kwargs.copy()
        no_sub = _kwargs.pop('no_sub', False)
        if no_sub:
            rv = ' '.join([str(self.call_eval(i, caller, **kwargs)) for i in self.objects])
        else:
            rv = str(self.call_eval(self.obj, caller, **kwargs)).format(
                *[self.call_eval(i, caller, **kwargs) for i in self.objects[1:]], **_kwargs)
        return rv


class Eval(AbstractAction):

    """Execute python command given as a string with eval (or exec).

    Usage::

        Eval("print time.{param}()", pre_exec="import time", param="time")

    First argument: python command to be evaluated. If it can be evaluated by
    eval() then return value is the evaluated value. Otherwise, exec() is used and True
    is returned.

    :param str pre_exec: pre-execution string. For example necessary import commands.
    :param dict namespace: Namespace. Defaults to locals() in :mod:`.builtin_callables`.

    Optionally, other keyword arguments can be given, and they are replaced in the first argument
    by format().

    See also (and prefer using): :class:`.Func`
    """

    def call(self, caller, **kwargs):
        if not caller:
            return
        _kwargs = self._kwargs.copy()
        namespace = _kwargs.pop('namespace', locals())
        pre_exec = _kwargs.pop('pre_exec', '')

        if pre_exec:
            exec(pre_exec.format(**self._kwargs), namespace)
        try:
            return eval(self.obj.format(**self._kwargs), namespace)
        except SyntaxError:
            exec(self.obj.format(**self._kwargs), namespace)
            return True


class Exec(Eval):

    """Synonym to :class:`.Eval`"""


class GetService(AbstractAction):
    """
    Get service by name and number.

    Usage::

        GetService(name)
        GetService(name, number)

    Usage examples::

        GetService('WebService')
        GetService('WebService', 1)

    """
    def call(self, caller, **kwargs):
        name = self._args[0]
        num = self._args[1] if len(self._args) > 1 else 0
        return self.system.services_by_name[name][num]


class ReloadService(AbstractAction):
    """
    Reload given service.

    Usage::

        ReloadService(name, number)
        ReloadService(name)

    Usage examples::

        ReloadService('WebService', 0)
        ReloadService('ArduinoService')

    """
    def call(self, caller, **kwargs):
        name = self._args[0]
        num = self._args[1] if len(self._args) > 1 else 0
        return self.system.services_by_name[name][num].reload()


class Shell(AbstractAction):

    """
        Execute shell command and return string value

        :param bool no_wait: if True, execute shell command in new thread and return pid
        :param bool output: if True, execute will return the output written to stdout by shell command. By default,
                            execution status (integer) is returned.
        :param str input: if given, input is passed to stdin of the given shell command.

        Usage examples::

            Shell('/bin/echo test', output=True) # returns 'test'
            Shell('mplayer something.mp3', no_wait=True) # returns PID of mplayer
                                                         # process that keeps running
            Shell('/bin/cat', input='test', output=True) # returns 'test'.
    """

    def call(self, caller, **kwargs):
        cmd = self.call_eval(self.obj, caller, **kwargs)
        input = self._kwargs.get('input', None)
        if input:
            input = bytes(self.call_eval(input, caller, **kwargs), 'utf-8')
        if not caller:
            return
        try:
            process = subprocess.Popen(
                cmd, executable='bash', shell=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
            self.logger.debug('Shell: cmd "%s", pid %s', cmd, process.pid)
            if self._kwargs.get('no_wait', False):
                thread_start(lambda: process.communicate(input))
                return process.pid
            else:
                out, err = process.communicate(input)
                retcode = process.poll()
                if self._kwargs.get('output', False):
                    return out.decode('utf-8')
                else:
                    return retcode
        except Exception as e:
            self.system.logger.error('Error %s in %s, cmd: %s', e, self, cmd)
        return -1


class SetStatus(AbstractAction):

    """
        Set sensor or actuator value

        Usage::

            SetStatus(target, source)
            # sets status of target to the status of source.
            SetStatus(target, source, Force=True)
            # sets status to hardware level even if it is not changed
            SetStatus([actuator1, actuator2], [sensor1, sensor2])
            # sets status of actuator 1 to status of sensor1 and
            # status of actuator2 to status of sensor2.
    """

    def call(self, caller=None, trigger=None, **kwargs):
        if not caller:
            return
        force = self._kwargs.get('force', False)
        values = self.call_eval(self.value, caller, **kwargs)
        objs = self.name_to_system_object(self.obj)
        if isinstance(objs, AbstractCallable):
            objs = self.call_eval(objs, caller, return_value=False, **kwargs)
        if not is_iterable(objs):
            objs = [objs]
        objs = list(objs)
        if not is_iterable(values):
            values = [values] * len(objs)
        values = list(values)
        if len(objs) != len(values):
            raise RuntimeError('SetStatus: length of targets should be equal to '
                                          'length of sources: %s != %s' % (len(objs), len(values)))

        self.system.logger.debug('SetStatus: objs %s, rv: %s', objs, values)
        for idx, obj in enumerate(objs):
            value = self.call_eval(values[idx], caller, **kwargs)
            obj = self.call_eval(obj, caller, return_value=False, trigger=trigger, **kwargs)
            self.system.logger.debug('SetStatus(%s, %s) by %s', obj, value, caller)
            try:
                obj.set_status(value, origin=caller, force=force)
            except ValueError as e:
                self.system.logger.error(
                    'Trying to set invalid status %s of type %s (by %s). Error: %s', value, type(value), caller, e)
            except AttributeError as e:
                self.system.logger.error(
                    'Trying to set status of invalid object %s of type %s, by %s. Error: %s', obj, type(obj), caller, e)
        return True

    def _give_triggers(self):
        return self.value

    def _give_targets(self):
        return self.obj


class SetAttr(AbstractAction):

    """Set object's attributes

       Usage::

            SetAttr(obj, attr=value, attr2=value2)
            # performs setattr(obj, attr, value) and setattr(obj, attr2, value2).
    """

    def __init__(self, obj, **kwargs):
        super(SetAttr, self).__init__(obj, **kwargs)

    def call(self, caller, **kwargs):
        if not caller:
            return
        obj = self.obj
        for attr, val in list(self._kwargs.items()):
            val = self.call_eval(val, caller, **kwargs)
            setattr(obj, attr, val)
        return True

    def _give_triggers(self):
        return list(self._kwargs.values())

    def _give_targets(self):
        return self.obj


class Changed(AbstractCallable):

    """
        Is value changed since evaluated last time? If this is the first time this Callable
        is called (i.e. comparison to last value cannot be made), return True.

        Usage::

            Changed(sensor1)
    """
    _lastval = Any(transient=True)

    def call(self, caller, **kwargs):
        if not caller:
            return
        with self._lock:
            newval = self.call_eval(self.obj, caller, **kwargs)
            rval = self._lastval != newval
            self.logger.debug("Changed(%s) last:%s new:%s: rval:%s", caller, self._lastval, newval, rval)
            self._lastval = newval
            return rval

    def _give_targets(self):
        return None


class Swap(AbstractAction):

    """
        Swap sensor or BinaryActuator status (False to True and True to False)

        Usage::

            Swap(actuator1)
    """

    def call(self, caller, trigger=None, **kwargs):
        if not caller:
            return
        rv = not (self.call_eval(self.obj, caller, **kwargs))
        self.logger.debug("Swap(%s): rv:%s", caller, rv)
        if trigger == self.obj:
            self.logger.debug('Swap: ignoring setting trigger value')
            return False
        self.obj.set_status(rv, caller)
        return rv

    def _give_triggers(self):
        return None


class AbstractRunner(AbstractAction):

    """ Abstract baseclass for Callables that are used primarily to run other Actions """


class Run(AbstractRunner):

    """Run specified Callables one at time. Return always True.

    Usage::

        Run(callable1, callable2, ...)
    """

    def call(self, caller=None, **kwargs):
        if not caller:
            return
        for i in self.objects:
            self.call_eval(i, caller, **kwargs)
        return True


class Delay(AbstractRunner):

    """Execute commands delayed by time (in seconds) in separate thread

    Usage::

        Delay(delay_in_seconds, action)
    """

    @property
    def delay(self):
        return self._args[0]

    @property
    def objects(self):
        return self._args[1:]

    def call(self, caller, **kwargs):
        if not caller:
            return

        with self._lock:
            state = self.get_state(caller)

            timers = state.get_or_create('timers', [])

            self.logger.info("Scheduling %s", self)
            delay = self.call_eval(self.delay, caller, **kwargs)
            timer = threading.Timer(delay, None)
            timer.function = threaded(self._run, caller, timer, **kwargs)
            time_after_delay = datetime.datetime.now() + datetime.timedelta(seconds=delay)
            timer.name = "Timer for %s timed at %s (%d sek)" % (self, time_after_delay, delay)
            timer.start()
            timers.append(timer)

    def cancel(self, caller):
        with self._lock:
            state = self.get_state(caller)
            timers = state.get('timers', [])

            for timer in timers:
                if timer.is_alive():
                    self.logger.info("Cancelling %s", self)
                    timer.cancel()
            self.del_state(caller)

        super(Delay, self).cancel(caller)

    def _run(self, caller, timer, **kwargs):
        self.logger.info("Time is up, running %s", self)
        for i in self.objects:
            if not caller in self.state:
                # cancelled
                return
            self.call_eval(i, caller, **kwargs)

        with self._lock:
            if caller in self.state: # if not cancelled
                self.get_state(caller).timers.remove(timer)
        return True


class Threaded(Delay):

    """Execute commands in a single thread (in order)

    Usage::

        Threaded(action)

    """

    def __init__(self, *args, **kwargs):
        super(Threaded, self).__init__(0, *args, **kwargs)


class If(AbstractCallable):

    """Basic If statement

    Usage::

        If(x, y, z) # if x, then run y, z, where x, y, and z are Callables or StatusObjects
        If(x, y)

    """

    def call(self, caller=None, **kwargs):
        if self.call_eval(self.objects[0], caller, **kwargs):
            objs = self.objects[1:]
            if len(objs) > 1:
                for i in objs:
                    self.call_eval(i, caller, **kwargs)
                return True
            else:
                return self.call_eval(self.objects[1], caller, **kwargs)
        else:
            return False

    def _give_triggers(self):
        return self.objects[1:]

    def _give_targets(self):
        return self.objects[1:]


class IfElse(AbstractCallable):

    """
    Basic if - then - else statement

    Usage::

        IfElse(x, y, z) # if x, then run y, else run z, where x, y,
                        # and z are Callables or StatusObjects
        IfElse(x, y)
    """

    def call(self, caller=None, **kwargs):
        if self.call_eval(self.objects[0], caller, **kwargs):
            return self.call_eval(self.objects[1], caller, **kwargs)
        if len(self.objects) > 2:
            return self.call_eval(self.objects[2], caller, **kwargs)
        else:
            return False

    def _give_triggers(self):
        return self.objects[1:]

    def _give_targets(self):
        return self.objects[1:]


class Switch(AbstractCallable):

    """
    Basic switch - case statement.

    Two alternative usages:

    * First argument switch criterion (integer-valued) and others are cases **OR**
    * First argument is switch criterion and second argument is dictionary that contains all possible
      cases as keys and related actions as their values.

    Usage::

        Switch(criterion, choice1, choice2...) # where criteria is integer-valued
                                               # (Callable or StatusObject etc.)
                                               # and choice1, 2... are Callables.

        Switch(criterion, {'value1': callable1, 'value2': 'callable2'})
    """

    def call(self, caller=None, **kwargs):
        sel = self.call_eval(self.objects[0], caller, **kwargs)
        if isinstance(self.value, dict):
            return self.call_eval(self.value.get(sel, None), caller, **kwargs)
        else:
            return self.call_eval(self.objects[sel + 1], caller, **kwargs)

    def _give_targets(self):
        return deep_iterate(self.objects[1:])

    def _give_triggers(self):
        return deep_iterate(self.objects[1:])


class TryExcept(AbstractRunner):

    """Try returning `x`, but if exception occurs in the value evaluation, then return `y`.

    Usage::

        Try(x, y) # where x and y are Callables or StatusObjects etc.
    """

    def call(self, caller=None, **kwargs):
        try:
            return self.call_eval(self.objects[0], caller, **kwargs)
        except:
            return self.call_eval(self.objects[1], caller, **kwargs)


class AbstractMathematical(AbstractCallable):

    def _give_targets(self):
        return None


class Min(AbstractMathematical):

    """
    Give minimum number of given objects.

    Usage::

        Min(x, y, z...)
        # where x,y,z are anything that can be
        # evaluated as number (Callables, Statusobjects etc).
     """

    def call(self, caller=None, **kwargs):
        val = float("inf")
        for i in self.objects:
            val = min(val, self.call_eval(i, caller, **kwargs))
        return val


class Max(AbstractMathematical):

    """Give maximum number of given objects

    Usage::

        Max(x, y, z...)
        # where x,y,z are anything that can be
        # evaluated as number (Callables, Statusobjects etc).
    """

    def call(self, caller=None, **kwargs):
        val = -float("inf")
        for i in self.objects:
            val = max(val, self.call_eval(i, caller, **kwargs))
        return val


class Sum(AbstractMathematical):

    """Give sum of given objects

    Usage::

        Sum(x, y, z...)
        # where x,y,z are anything that can be
        # evaluated as number (Callables, Statusobjects etc).
    """

    def call(self, caller=None, **kwargs):
        _sum = 0.0
        for i in self.objects:
            _sum += self.call_eval(i, caller, **kwargs)
        return _sum


class Product(AbstractMathematical):

    """Give product of given objects

    Usage::

        Product(x, y, z...)
        # where x,y,z are anything that can be
        # evaluated as number (Callables, Statusobjects etc).

    """

    def call(self, caller=None, **kwargs):
        _sum = 1.0
        for i in self.objects:
            _sum *= self.call_eval(i, caller, **kwargs)
        return _sum


class Mult(Product):

    """Synonym of Product"""


class Division(AbstractMathematical):

    """Give division of given objects

    Usage::

        Division(x, y)
        # where x,y are anything that can be
        # evaluated as number (Callables, Statusobjects etc).

    """

    def call(self, caller=None, **kwargs):
        if len(self.objects) != 2:
            raise RuntimeError('Wrong amount of arguments')

        obj1, obj2 = self.objects
        _val1 = self.call_eval(obj1, caller, **kwargs)
        _val2 = self.call_eval(obj2, caller, **kwargs)
        return _val1 / _val2


class Div(Division):

    """ Synonym of Division"""


class Add(Sum):

    """Synonym of Sum """


class AbstractLogical(AbstractMathematical):

    """Abstract class for logic operations (:class:`.And`, :class:`.Or` etc.) """


class Anything(AbstractLogical):

    """ Condition which gives `True` always

    Usage::

        Anything(x,y,z...)
    """

    def call(self, caller=None, **kwargs):
        return True


class Or(AbstractLogical):

    """ Or condition

    Usage::

        Or(x,y,z...) # gives truth value of x or y or z or ,,,
    """

    def call(self, caller=None, **kwargs):
        def _or(list):
            for i in list:
                val = self.call_eval(i, caller, **kwargs)
                if is_iterable(val):
                    val = _or(val)
                if val:
                    return True
            return False

        return _or(self.objects)


class And(AbstractLogical):

    """And condition

    Usage::

        And(x,y,z...) # gives truth value of x and y and z and ...

    """

    def call(self, caller=None, **kwargs):
        def _and(list):
            for i in list:
                val = self.call_eval(i, caller, **kwargs)
                if is_iterable(val):
                    val = _and(val)
                if not val:
                    return False
            return True
        return _and(self.objects)


class Neg(AbstractLogical):

    """Give negative of specified callable (minus sign)

    Usage::

        Neg(x) # returns -x
    """

    def call(self, caller=None, **kwargs):
        if len(self.objects) != 1:
            raise RuntimeError('Too many arguments')
        return -self.call_eval(self.obj, caller, **kwargs)


class Inverse(AbstractLogical):

    """Give inverse of specified callable (1/something)

    Usage::

        Inv(x) # returns 1/x
    """

    def call(self, caller=None, **kwargs):
        if len(self.objects) != 1:
            raise RuntimeError('Too many arguments')
        return 1./self.call_eval(self.obj, caller, **kwargs)


class Inv(Inverse):

    """ Synonym of Inverse """


class Not(AbstractLogical):

    """Give negation of specified object

    Usage::

        Not(x) # returns not x
    """

    def call(self, caller=None, **kwargs):
        if len(self.objects) != 1:
            raise RuntimeError('Too many arguments')
        return not self.call_eval(self.obj, caller, **kwargs)


class Equal(AbstractLogical):

    """Equality condition, i.e. is x == y

    Usage::

        Equal(x, y) # returns truth value of x == y
    """

    def call(self, caller=None, **kwargs):
        return self.call_eval(self.obj, caller, **kwargs) == self.call_eval(self.value, caller, **kwargs)


class Less(AbstractLogical):

    """Condition: is x < y

    Usage::

        Less(x,y) # returns truth value of x < y
    """

    def call(self, caller=None, **kwargs):
        a = self.call_eval(self.obj, caller, **kwargs)
        b = self.call_eval(self.value, caller, **kwargs)
        try:
            rv = a < b
        except TypeError:
            rv = False
        return rv

class More(AbstractLogical):

    """Condition: is x > y

    Usage::

        More(x,y) # returns truth value of x > y

    """

    def call(self, caller=None, **kwargs):
        a = self.call_eval(self.obj, caller, **kwargs)
        b = self.call_eval(self.value, caller, **kwargs)
        try:
            rv = a > b
        except TypeError:
            rv = False
        return rv


class Value(AbstractLogical):

    """Give specified value

    Usage::

        Value(x) # returns value of x. Used to convert StatusObject into Callable,
                 # for example, if StatusObject status needs to be used directly
                 # as a condition of Program condition attributes.

    """
    _args = CList

    def call(self, caller=None, **kwargs):
        return self.call_eval(self.obj, caller, **kwargs)


class AbstractQuery(AbstractCallable):

    """
    Baseclass for query type of Callables, i.e. those that return
    set of objects from system based on given conditions.
    """


class ReprObject(object):

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return self.name


class OfType(AbstractQuery):

    """
    Gives all objects of given type that are found in System

    Usage & example::

        OfType(type, **kwargs)
        OfType(AbstractActuator, exclude=['actuator1', 'actuator2'])
        # returns all actuators in system, except those named 'actuator1' and 'actuator2'.

    :param list exclude: list of instances to be excluded from the returned list.

    """

    system_objects_changed = Event
    triggers = Property(trait=Set(trait=StatusObject),
                        depends_on='on_setup_callable, _kwargs_items, _args_items, system_objects_changed')

    targets = Property(trait=Set(trait=StatusObject),
                       depends_on='on_setup_callable, _kwargs, _kwargs_items, _args, _args_items, '
                                  'system_objects_changed')

    setup_complete = CBool(transient=True)

    @on_trait_change('system')
    def set_setup_complete(self, name, new):
        if name == 'system':
            self.system.on_trait_change(self.set_setup_complete, 'post_init_trigger')
        if name == 'post_init_trigger':
            self.setup_complete = True

    @on_trait_change('system.objects, system.objects_items, setup_complete')
    def trigger_system_objects_changed(self):
        if self.setup_complete:
            self.system_objects_changed = 1

    def _both(self):
        if not self.system:
            return set()
        exclude = set(self._kwargs.get('exclude', []))  # exclude must be list type
        return {i for i in self.system.objects if isinstance(i, tuple(self.objects))} - exclude

    @cached_property
    def _get_triggers(self):
        if self._kwargs.get('type', 'both') in ['triggers', 'both']:
            return self._both()
        return set()

    @cached_property
    def _get_targets(self):
        if self._kwargs.get('type', 'both') in ['targets', 'both']:
            return self._both()
        return set()

    def call(self, caller=None, **kwargs):
        return self.targets

    def give_str(self):
        args = [ReprObject(i.__name__) for i in self.objects]
        return self._give_str(args, self._kwargs)

    def give_str_indented(self, tags=False):
        args = [ReprObject(i.__name__) for i in self.objects]
        return self._give_str_indented(args, self._kwargs, tags)


class RegexSearch(AbstractCallable):

    """
    Scan through string looking for a match to the pattern.
    Return matched parts of string by :func:`re.search`.

    :param int group: Match group can be chosen by group number.

    Usage & examples::

        RegexSearch(match_string, content_to_search, **kwargs)

        RegexSearch(r'(\d*)(\w*)', '12test')          # returns '12'
        RegexSearch(r'(\d*)(\w*)', '12test', group=2) # returns 'test'
        RegexSearch(r'testasfd', 'test')              # returns ''

    .. tip::
        More examples in unit tests

    """

    def call(self, caller=None, **kwargs):
        pattern = str(self.call_eval(self.obj, caller, **kwargs))
        searchstring = str(self.call_eval(self.value, caller, **kwargs))
        match = re.search(pattern, searchstring, re.MULTILINE)
        if match:
            return match.group(self._kwargs.get('group', 1))
        else:
            return ''


class RegexMatch(AbstractCallable):

    """
    Try to apply the pattern at the start of the string.
    Return matched parts of string by :func:`re.match`.

    :param int group: Match group can be chosen by group number.

    Usage & examples from unit tests::

        RegexMatch(match_string, content_to_search, **kwargs)

        RegexMatch(r'heptest', 'heptest')                # returns 'heptest'
        RegexMatch(r'heptest1', 'heptest')               # returns ''
        RegexMatch(r'(hep)te(st1)', 'heptest1', group=1) # returns 'hep'
        RegexMatch(r'(hep)te(st1)', 'heptest1', group=2) # returns 'st1'

    .. tip::
        More examples in unit tests

    """

    def call(self, caller=None, **kwargs):
        pattern = self.call_eval(self.obj, caller, **kwargs)
        searchstring = self.call_eval(self.value, caller, **kwargs)
        match = re.match(pattern, searchstring, re.MULTILINE)
        if match:
            return match.group(self._kwargs.get('group', 0))
        else:
            return ''



class RemoteFunc(AbstractCallable):

    """
        Evaluate remote function via XMLRPC.

        Usage::

            RemoteFunc('host', 'funcname', *args, **kwargs)
    """
    _cached_server = Any(transient=True)

    def call(self, caller, **kwargs):
        try:
            if self._cached_server is None:
                host = self.call_eval(self.obj, caller, **kwargs)
                self._cached_server = server = xmlrpc.client.ServerProxy(host)
            else:
                server = self._cached_server

            funcname = self.call_eval(self.value, caller, **kwargs)
            args = [self.call_eval(o, caller, **kwargs) for o in self.objects[2:]]
            _kwargs = {k: self.call_eval(v, caller, **kwargs) for k, v in list(self._kwargs.items())}
            try:
                return getattr(server, funcname)(*args, **_kwargs)
            except (xmlrpc.client.Fault, HTTPException) as e:
                self.logger.exception(
                    'Exception occurred in remote function call (%s,%s)(*%s, **%s), error: %s', server, funcname, args, _kwargs, e)
        except (socket.gaierror, IOError, xmlrpc.client.Fault) as e:
            self.logger.error('Could call remote function, error: %s', e)


class WaitUntil(AbstractRunner):

    """
        Wait until sensor/actuator/callable status changes to True and then execute commands.
        WaitUntil will return immediately and only execute specified actions
        after criteria is fullfilled.

        Usage::

            WaitUntil(sensor_or_callable, Action1, Action2, etc)

        .. note::
            No triggers are collected from WaitUntil

    """

    def call(self, caller, **kwargs):
        obj = self.name_to_system_object(self.obj)

        def callback(**kwargs2):
            nocallback = kwargs2.get('nocallback', False)
            callbacks = self.get_state(caller).callbacks if not nocallback else []

            if obj.status and (nocallback or (callbacks and callback in callbacks)):
                for i in self.objects[1:]:
                    self.call_eval(i, caller, **kwargs)
                if not nocallback:
                    self.logger.debug('Removing executed callback')
                    with self._lock:
                        obj.on_trait_change(callback, 'status', remove=True)
                        callbacks.remove(callback)
        if not obj.status:
            self.logger.debug('Adding callback')
            obj.on_trait_change(callback, 'status')
            with self._lock:
                callbacks = self.get_state(caller).get_or_create('callbacks', [])
                callbacks.append(callback)
        else:
            self.logger.debug('Calling directly, no callback')
            callback(nocallback=True)

    def cancel(self, caller):
        self.logger.debug('Cancelling callbacks')
        with self._lock:
            state = self.get_state(caller)
            callbacks = state.get('callbacks', [])
            for cb in callbacks:
                self.obj.on_trait_change(cb, 'status', remove=True)
            self.del_state(caller)
        super(WaitUntil, self).cancel(caller)

    def _give_triggers(self):
        return None


class While(AbstractRunner):

    """
        Executes commands (in thread) as long as criteria (sensor, actuator, callable status) remains true.
        Flushes worker queue between each iteration such that criteria is updated, if executed
        actions alter it.

        :param Callable do_after: given Callable is executed after while loop is finished.

        Usage & example::

            While(criteria, action1, action2, do_after=action3)

            # Example loop that runs actions 10 times. Assumes s=UserIntSensor()
            Run(
                SetStatus(s, 0),
                While(s < 10,
                    SetStatus(s, s+1),
                    other_actions
                )
            )

        .. note::
            While execution is performed in separate thread
        .. note::
            No triggers are collected from While


    """

    class ExitThread(Exception):
        pass

    def _run(self, caller, thread, **kwargs):
        self.system.flush()
        try:
            while self.call_eval(self.obj, caller, **kwargs):
                for i in self.objects[1:]:
                    if thread._cancel_while:
                        raise self.ExitThread
                    self.call_eval(i, caller, **kwargs)
                self.system.flush()
            do_after = self._kwargs.get('do_after')
            if do_after:
                self.call_eval(do_after, caller, **kwargs)
        except self.ExitThread:
            self.logger.debug('While exited via cancel')

        with self._lock:
            state = self.get_state(caller)
            state.threads.remove(thread)
            if not state.threads:
                self.logger.debug('Last thread, removing state')
                self.del_state(caller)

    def call(self, caller, **kwargs):
        with self._lock:
            state = self.get_state(caller)
            threads = state.get_or_create('threads', [])
            t = threading.Timer(0., None)
            t.function = threaded(self._run, caller, t, **kwargs)
            t.name = 'Thread for %s' % self
            t._cancel_while = False
            threads.append(t)
            t.start()
        return True

    def cancel(self, caller):
        self.logger.debug('Canceling While')
        with self._lock:
            state = self.get_state(caller)
            for t in state.get('threads', []):
                t._cancel_while = True
        super(While, self).cancel(caller)

    def _give_triggers(self):
        return None


class TriggeredBy(OnlyTriggers):

    """
        Return whether action was triggered by one of specified triggers or not

        If no arguments, return the trigger.

        Usage::

            TriggeredBy() # -> returns the trigger
            TriggeredBy(trig1, trig2...) #-> Returns if trigger is one of arguments
    """

    def call(self, caller=None, trigger=None, **kwargs):
        if self.objects:
            obj_eval = [self.name_to_system_object(obj) for obj in self.objects]
            return trigger in obj_eval
        else:
            return trigger


__all__ = get_modules_all(AbstractCallable, locals())



# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

from __future__ import unicode_literals
import logging
logger = logging.getLogger('automate')
import pytest
import logging

from automate.system import System


@pytest.fixture(params=[1])
def sysloader(request, tmpdir):
    filename = str(tmpdir.join('savefile.dmp'))
    if request.param:
        class Loader(object):

            def new_system(self, sys, *args, **kwargs):
                kwargs.setdefault('exclude_services', ['TextUIService'])
                self.sys = sys(*args, **kwargs)
                self.sys.flush()
                return self.sys
    else:
        class Loader(object):

            def new_system(self, sys, *args, **kwargs):
                kwargs.setdefault('exclude_services', ['TextUIService'])
                sys1 = sys(*args, **kwargs)
                sys1.flush()
                sys1.filename = filename
                sys1.save_state()
                sys1.cleanup()

                self.sys = System.load_or_create(filename, exclude_services=['TextUIService'])
                self.sys.flush()
                return self.sys

    loader = Loader()
    yield loader
    loader.sys.cleanup()


@pytest.fixture(autouse=True)
def check_log(request, caplog):
    formatter = logging.Formatter(fmt='%(threadName)s:%(asctime)s:%(name)s:%(levelname)s:%(message)s')
    caplog.handler.setFormatter(formatter)

    logger.info('UNITTEST %s', request)
    yield None
    if not getattr(caplog, 'error_ok', False):
        for i in caplog.records():
            if i.levelno >= logging.ERROR:
                raise Exception('Error: %s' % i.getMessage())



# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of automate-arduino.
#
# automate-arduino is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# automate-arduino is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with automate-arduino.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals
from traits.api import CInt, Instance, CBool, CFloat

from automate.actuators import FloatActuator
from automate.service import AbstractSystemService
from automate.statusobject import AbstractActuator


class AbstractArduinoActuator(AbstractActuator):

    """
        Abstract base class for Arduino actuators
    """

    #: Arduino device number (specify, if more than 1 devices configured in ArduinoService)
    dev = CInt(0)

    #: Arduino pin number
    pin = CInt

    _arduino = Instance(AbstractSystemService, transient=True)

    view = AbstractActuator.view + ["pin"]
    simple_view = AbstractActuator.simple_view + ["pin"]

    def setup(self, *args, **kwargs):
        super(AbstractArduinoActuator, self).setup(*args, **kwargs)
        self._arduino = self.system.request_service('ArduinoService')


class ArduinoDigitalActuator(AbstractArduinoActuator):

    """
        Boolean-valued actuator object for digital Arduino output pins
    """
    _status = CBool(transient=True)

    def setup(self, *args, **kwargs):
        super(ArduinoDigitalActuator, self).setup(*args, **kwargs)
        self._arduino.setup_digital(self.dev, self.pin)

    def _status_changed(self):
        self._arduino.change_digital(self.dev, self.pin, self._status)

    def cleanup(self):
        self._arduino.cleanup_digital_actuator(self.dev, self.pin)


class ArduinoServoActuator(AbstractArduinoActuator):

    """
        Float-valued actuator object for Arduino output pins that can be configured in Servo mode
        Status is servo angle (0-360).
    """

    _status = CFloat(transient=True)

    #: Minimum pulse time (in microseconds)
    min_pulse = CInt(544)

    #: Maximum pulse time (in microseconds)
    max_pulse = CInt(2400)

    def _min_pulse_changed(self):
        if self.traits_inited():
            self.setup()

    def _max_pulse_changed(self):
        if self.traits_inited():
            self.setup()

    def setup(self, *args, **kwargs):
        super(ArduinoServoActuator, self).setup(*args, **kwargs)
        self.logger.debug("setup_servo %s %s %s %s %s %s", self, self.dev, self.pin, self.min_pulse, self.max_pulse,
                          int(round(self._status)))
        self._arduino.setup_servo(self.dev, self.pin, self.min_pulse, self.max_pulse, int(round(self._status)))

    def _status_changed(self):
        self.logger.debug("change_servo %s %s %s", self.dev, self.pin, int(round(self._status)))
        self._arduino.change_digital(self.dev, self.pin, int(round(self._status)))

    def cleanup(self):
        self._arduino.cleanup_digital_actuator(self.dev, self.pin)

class ArduinoPWMActuator(FloatActuator, AbstractArduinoActuator):

    """
        Float-valued actuator object for Arduino output pins that can be configured in PWM mode
        Status is float between 0.0 and 1.0.
    """

    def setup(self, *args, **kwargs):
        super(ArduinoPWMActuator, self).setup(*args, **kwargs)
        self._arduino.setup_pwm(self.dev, self.pin)

    def _status_changed(self):
        self._arduino.change_digital(self.dev, self.pin, max(0., min(1., self._status)))

    def cleanup(self):
        self._arduino.cleanup_digital_actuator(self.dev, self.pin)


# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

from __future__ import absolute_import
from __future__ import unicode_literals
from past.builtins import basestring
from copy import copy
import logging
import re
import keyword
import threading
from collections import Iterable
from functools import wraps

from traits.api import CSet, HasStrictTraits, Unicode, TraitType


class AbstractStatusObject(object):

    """
        Only reason for this abstract baseclass is that we need to avoid import loops
        between statusobject.py and various other modules.
    """


class SystemNotReady(Exception):
    pass


class ExitException(Exception):
    pass


class SystemBase(HasStrictTraits):
    pass


class Group(object):
    pass


class Object(object):

    """
        Class to refer explicitly objects by name
    """

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name


class LogicStr(Unicode):

    def validate(self, obj, name, value):
        v = value
        f = obj.system.eval_in_system_namespace(v)
        if f is None:
            self.error(obj, name, value)
        return value


class NameOrSensorActuatorBaseTrait(TraitType):

    def validate(self, obj, name, value):
        from .statusobject import StatusObject
        v = value
        if isinstance(v, StatusObject):
            return v
        if isinstance(v, basestring):
            return obj.system.name_to_system_object(v)
        self.error(obj, name, value)
        return value


def threaded(func, *args, **kwargs):
    """ uses thread_init as a decorator-style """
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            import traceback
            logging.getLogger('automate.common').error(
                "Exception occurred in thread: %s\n Traceback: \n %s", e, traceback.format_exc())
            return False

    return lambda: wrapper(*args, **kwargs)


def thread_start(func, *args, **kwargs):
    tfunc = threaded(func, *args, **kwargs)
    t = threading.Thread(target=tfunc, name='thread_start')
    t.start()
    return t


def is_valid_variable_name(name):
    return re.match("^[_A-Za-z][_a-zA-Z0-9]*$", name) and not keyword.iskeyword(name)


def has_baseclass(v, s):
    try:
        return issubclass(v, s)
    except TypeError:
        return False


class Lock(object):

    """Lock object (similar to threading.Lock) that can print some debug information"""
    context = None

    def __init__(self, name="Unnamed lock", silent=False):
        self.logger = logging.getLogger('automate.lock')
        self.name = name
        self.lock = threading.Lock()
        self.context_lock = threading.Lock()
        self.silent = silent

    def acquire(self, wait=0):
        return self.__enter__()

    def release(self):
        return self.__exit__(None, None, None)

    def __enter__(self):
        import traceback
        if self.lock.acquire(False):
            pass
        else:
            self.logger.debug("WAITING for lock %s", self.name)
            with self.context_lock:
                if not self.silent and self.context:
                    current_context = traceback.format_stack()
                    self.logger.debug("Current context:\n %s", "".join(current_context))
                    self.logger.debug("The context, when lock was acquired\n %s", "".join(self.context))
            self.lock.acquire()
            self.logger.debug("OK, gone through %s", self.name)

        with self.context_lock:
            self.context = traceback.format_stack()

    def __exit__(self, type, value, tb):
        with self.context_lock:
            self.context = None
        self.lock.release()


class _nomutex(object):

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

nomutex = _nomutex()


class CompareMixin(object):

    """Comparison and other operators that are common for Sensors,
       Actuators and callables.py operations

       """

    def __mul__(self, obj):
        from . import callables
        return callables.Product(self, obj)

    def __add__(self, obj):
        from . import callables
        return callables.Sum(self, obj)

    def __sub__(self, obj):
        from . import callables
        return callables.Sum(self, -obj)

    def __neg__(self):
        from . import callables
        return callables.Neg(self)

    def __inv__(self):
        from . import callables
        return callables.Inv(self)

    def __truediv__(self, other):
        from . import callables
        return callables.Division(self, other)

    def __lt__(self, obj):
        from . import callables
        return callables.Less(self, obj)

    def __gt__(self, obj):
        from . import callables
        return callables.More(self, obj)


class TagSet(CSet):

    def validate(self, object, name, value):
        if isinstance(value, basestring):
            return set((i.strip() for i in value.split(',')))
        return super(TagSet, self).validate(object, name, value)


def is_iterable(y):
    if isinstance(y, basestring):
        return False
    return isinstance(y, Iterable)


class DictObject(dict):

    def __getattr__(self, item):
        if item in self:
            return self[item]

    def __setattr__(self, key, value):
        self[key] = value

    def get_or_create(self, k, default):
        if k in self:
            return self[k]
        else:
            self[k] = default
            return default


def deep_iterate(l):
    if is_iterable(l):
        l_list = l
        if isinstance(l, dict):
            l_list = list(l.values())
        for i in l_list:
            if is_iterable(i):
                for j in deep_iterate(i): # TODO: when dropping python 2 support, we can use yield from here (python 3.3+)
                    yield j
            else:
                yield i
    else:
        yield l


def get_modules_all(base_class, _locals):
    r_types = {}
    for k, v in list(copy(_locals).items()):
        if has_baseclass(v, base_class) and v is not base_class:
            r_types[k] = v

    return list(r_types.keys())



from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()
# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

import io

from traits.api import Str, Dict

from automate.service import AbstractUserService
from automate.program import DefaultProgram, ProgrammableSystemObject
from automate.statusobject import AbstractActuator, AbstractSensor


class PlantUMLService(AbstractUserService):

    """
        Provides UML diagrams of the system as SVG images. Used by WebService.

        PLantUMLService requires either PlantUML software (which is opensource software written in Java) to be
        installed locally (see http://plantuml.sourceforge.net/) or it is possible to use online service of plantuml.com
        In addition you need python package :mod:`plantuml` (available via PYPI).
    """

    #: URL of PlantUML Java Service. To use PlantUML online service, set this to 'http://www.plantuml.com/plantuml/svg/'
    url = Str()

    #: Arrow colors as HTML codes stored as a dictionary with keys:
    #: controlled_target, active_target, inactive_target, trigger
    arrow_colors = Dict(
        dict(controlled_target='#FF0000', active_target='#0000FF', inactive_target='#4C4C4C', trigger='#009933'))

    #: Background colors as HTML codes, stored as a dictionary with keys: program, actuator, sensor
    background_colors = Dict(dict(program='#FFFFCC', actuator='#FFCCFF', sensor='#CCFFCC'))

    def write_puml(self, filename=''):
        """
            Writes PUML from the system. If filename is given, stores result in the file.
            Otherwise returns result as a string.
        """
        def get_type(o):
            type = 'program'
            if isinstance(o, AbstractSensor):
                type = 'sensor'
            elif isinstance(o, AbstractActuator):
                type = 'actuator'
            return type

        if filename:
            s = open(filename, 'w')
        else:
            s = io.StringIO()
        s.write('@startuml\n')
        s.write('skinparam state {\n')
        for k, v in list(self.background_colors.items()):
            s.write('BackGroundColor<<%s>> %s\n' % (k, v))
        s.write('}\n')

        for o in self.system.objects:
            if isinstance(o, DefaultProgram) or o.hide_in_uml:
                continue

            if isinstance(o, ProgrammableSystemObject):
                s.write('state "%s" as %s <<%s>>\n' % (o, o, get_type(o)))

                s.write('%s: %s\n' % (o, o.class_name))
                if isinstance(o, AbstractActuator):
                    for p in reversed(o.program_stack):
                        s.write('%s: %s :: %s\n' % (o, p, o.program_status.get(p, '-')))
                elif hasattr(o, 'status'):
                    s.write('%s: Status: %s\n' % (o, o.status))
                if getattr(o, 'is_program', False):
                    s.write('%s: Priority: %s\n' % (o, o.priority))

                for t in o.actual_triggers:
                    if isinstance(t, DefaultProgram) or t.hide_in_uml:
                        continue
                    s.write('%s -[%s]-> %s\n' % (t, self.arrow_colors['trigger'], o))
                for t in o.actual_targets:
                    if t.hide_in_uml:
                        continue
                    if o.active:
                        color = 'active_target'
                    else:
                        color = 'inactive_target'
                    if getattr(t, 'program', None) == o:
                        color = 'controlled_target'

                    s.write('%s -[%s]-> %s\n' % (o, self.arrow_colors[color], t))
        s.write('@enduml\n')
        if filename:
            s.close()
        else:
            return s.getvalue()

    def write_svg(self):
        """
            Returns PUML from the system as a SVG image. Requires plantuml library.
        """
        import plantuml
        puml = self.write_puml()
        server = plantuml.PlantUML(url=self.url)
        svg = server.processes(puml)
        return svg



# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of automate-wsgi.
#
# automate-wsgi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# automate-wsgi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with automate-wsgi.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals
import threading
import socket

import tornado
import tornado.wsgi
import tornado.httpserver
import tornado.ioloop
import tornado.web
import tornado.websocket

from traits.api import Instance, Int, CStr, Dict, Str

from automate.common import threaded
from automate.service import AbstractUserService

web_thread = None


class TornadoService(AbstractUserService):
    """
    Abstract service that provides HTTP server for WSGI applications.
    """

    #: Which ip address to listen. Use ``0.0.0.0`` (default) to listen to all local networking interfaces.
    http_ipaddr = CStr("0.0.0.0")

    #: HTTP (or HTTPS if using SSL) port to listen
    http_port = Int(3000)

    #: Path to ssl certificate file. If set, SSL will be used.
    #:
    #: .. tip::
    #:
    #:   You may use script scripts/generate_selfsigned_certificate.sh to generate a
    #:   self-signed openssl certificate.
    ssl_certificate = CStr

    #: Path to ssl private key file
    ssl_private_key = CStr

    #: Number of listener threads to spawn
    num_threads = Int(5)

    #: Extra static dirs you want to serve. Example::
    #:
    #:    static_dirs = {'/my_static/(.*)': '/path/to/my_static'}
    static_dirs = Dict(key_trait=Str, value_trait=Str)

    _http_server = Instance(tornado.httpserver.TCPServer)

    @property
    def is_alive(self):
        return bool(self._http_server)

    def get_wsgi_application(self):
        """
            Get WSGI function. Implement this in subclasses.
        """
        raise NotImplementedError

    def get_websocket(self):
        return None

    def get_filehandler_class(self):
        return tornado.web.StaticFileHandler

    def get_tornado_handlers(self):
        tornado_handlers = []
        websocket = self.get_websocket()
        if websocket:
            tornado_handlers.append(('/socket', websocket))

        for entrypoint, path in self.static_dirs.items():
            tornado_handlers.append((entrypoint, self.get_filehandler_class(), {'path': path}))

        wsgi_app = self.get_wsgi_application()

        if wsgi_app:
            wsgi_container = tornado.wsgi.WSGIContainer(wsgi_app)
            tornado_handlers.append(('.*', tornado.web.FallbackHandler, dict(fallback=wsgi_container)))
        return tornado_handlers

    def setup(self):
        if self.is_alive:
            self.logger.debug('Server is already running, no need to start new')

        tornado_app = tornado.web.Application(self.get_tornado_handlers())

        if self.ssl_certificate and self.ssl_private_key:
            ssl_options = {
                "certfile": self.ssl_certificate,
                "keyfile": self.ssl_private_key,
            }
        else:
            ssl_options = None

        self._http_server = tornado.httpserver.HTTPServer(tornado_app, ssl_options=ssl_options)

        try:
            self._http_server.listen(self.http_port, self.http_ipaddr)
        except socket.error as e:
            self.logger.error('Could not start server: %s', e)
            self._http_server = None
            return

        self.start_ioloop()

    def start_ioloop(self):
        global web_thread
        ioloop = tornado.ioloop.IOLoop.instance()
        if not ioloop._running:
            web_thread = threading.Thread(target=threaded(ioloop.start),
                                          name="%s::%s" % (self.system.name, self.__class__.__name__))
            web_thread.start()

    def cleanup(self):
        if self.is_alive:
            tornado.ioloop.IOLoop.instance().stop()
            self._http_server.stop()
            self._http_server = None
            web_thread.join()



from __future__ import unicode_literals
from builtins import range
# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

from collections import deque

import pytest
import time

from automate import *
from automate import Group
from automate.program import Program

logger = logging.getLogger('automate.tests')

DELAYTIME = 1.0
SLEEPTIME = 1.7 * DELAYTIME
LOW = 1.0
HIGH = 100.0
AMOUNT = 20


@pytest.fixture(params=[0, 1, 2])
def mysys_old(request):
    if request.param == 0:
        class _mysys(System):
            act = FloatActuator(default=LOW, debug=True)
            slaveact = FloatActuator(default=LOW, debug=True, slave=True)
            slavebinact = BoolActuator(default=False, debug=True, slave=True)
            sens = UserBoolSensor(default=False, debug=True)
            s1 = UserBoolSensor(default=False, debug=True)
            s2 = UserFloatSensor(default=HIGH)
            slaveprog = Program(active_condition=Value(s1),
                                on_update=SetStatus(slaveact, s2))
            p2 = Program(active_condition=Value(s1),
                         on_activate=SetStatus(act, HIGH),
                         priority=2,
                         )
            prog = Program(
                active_condition=Value(sens),
                on_activate=SetStatus(act, HIGH),
            )
    elif request.param == 1:
        class _mysys(System):
            act = FloatActuator(default=LOW, debug=True)
            slaveact = FloatActuator(default=LOW, debug=True, slave=True)
            slavebinact = BoolActuator(default=False, debug=True, slave=True)
            sens = UserBoolSensor(default=False, debug=True)
            s1 = UserBoolSensor(default=False, debug=True)
            s2 = UserFloatSensor(default=HIGH)

            slaveprog = Program(active_condition=Value('s1'),
                                on_update=SetStatus('slaveact', 's2'))
            p2 = Program(active_condition=Value('s1'),
                         on_activate=SetStatus('act', HIGH),
                         priority=2,
                         )
            prog = Program(
                active_condition=Value('sens'),
                on_activate=SetStatus('act', HIGH),
            )
    else:
        class _mysys(System):

            class group1(Group):
                act = FloatActuator(default=LOW, debug=True)
                slaveact = FloatActuator(default=LOW, debug=True, slave=True)
                slavebinact = BoolActuator(default=False, debug=True, slave=True)
                sens = UserBoolSensor(default=False, debug=True)

                class group3(Group):
                    s1 = UserBoolSensor(default=False, debug=True)
                    s2 = UserFloatSensor(default=HIGH)

            class group2(Group):
                slaveprog = Program(active_condition=Value('s1'),
                                    on_update=SetStatus('slaveact', 's2'))
                p2 = Program(active_condition=Value('s1'),
                             on_activate=SetStatus('act', HIGH),
                             priority=2,
                             )
                prog = Program(
                    active_condition=Value('sens'),
                    on_activate=SetStatus('act', HIGH),
                )

    yield _mysys
#    s = _mysys(exclude_services=['TextUiService'])
#    s.namespace['defprog'] = s.act.program
#    s.flush()
#    s.logger.info('UNITTEST %s', request)
#    yield s
#    s.cleanup()


@pytest.fixture
def mysys(mysys_old, sysloader):
    s = sysloader.new_system(mysys_old)
    s.namespace['defprog'] = s.act.program
    yield s

#@pytest.fixture
# def mysys(mysys_old):
#    s = mysys_old(exclude_services=['TextUiService'])
#    s.namespace['defprog'] = s.act.program
#    yield s
#    s.cleanup()


@pytest.fixture
def mysys_groups():
    class _mysys(System):

        class group1(Group):
            s1 = UserFloatSensor(default=HIGH)

        class group2(Group):
            s2 = UserFloatSensor(default=HIGH)

            class group3(Group):
                s3 = UserFloatSensor()

        act = FloatActuator(default=LOW, debug=True)
        p2 = Program(active_condition=Value('s1'),
                     on_activate=SetStatus('act', HIGH),
                     priority=2,
                     )

    s = _mysys(exclude_services=['TextUIService'])
    s.flush()
    yield s
    s.cleanup()


def test_groups(mysys_groups):
    s = mysys_groups
    assert {'group:group1'} == s.s1.tags
    assert {'group:group2'} == s.s2.tags
    assert s.s3.tags == {'group:group2', 'group:group3'}  # 'group:group2.group3'}
    assert s.p2.active_condition.obj is s.s1
    assert s.p2.on_activate.obj is s.act
    assert s.p2.on_activate.value is HIGH


def test_samename(caplog):
    class _mysys(System):

        class group1(Group):
            s1 = UserFloatSensor(default=HIGH)

        class group2(Group):
            s1 = UserFloatSensor(default=HIGH)
    with pytest.raises(NameError):
        s = _mysys()
    caplog.error_ok = True


def test_slave_actuator(mysys):

    with pytest.raises(ValueError):
        mysys.act.status = HIGH

    assert mysys.act._status == LOW
    a = mysys.slaveact
    assert a._status == LOW
    a.status = HIGH
    mysys.flush()
    assert a._status == HIGH
    a.set_status(LOW)
    mysys.flush()
    assert a._status == LOW
    mysys.s1.set_status(True)
    mysys.flush()
    assert a._status == HIGH
    mysys.s2.set_status(LOW)  # slaveprog should set status of a to LOW
    mysys.flush()
    assert a._status == LOW
    assert mysys.slaveprog.active
    assert a in mysys.slaveprog.actual_targets
    mysys.s1.set_status(False)
    mysys.flush()
    assert not mysys.slaveprog.active
    # if program is deactivated, slave actuator status is not changed
    # I'm not sure if it should be changed back to default value, but
    # not it's like that
    assert a._status == LOW


def test_default_statuses(mysys):
    assert mysys.sens._status == False
    assert mysys.s1._status == False
    assert mysys.act._status == LOW
    assert mysys.slaveact._status == LOW


def test_program_activation(mysys):
    assert mysys.act.program == mysys.defprog
    assert len(mysys.act.program_stack) == 1
    assert mysys.act.status == LOW
    assert mysys.act.get_program_status(mysys.prog) == None
    assert mysys.prog.active == False
    assert mysys.act.program.name == 'dp_act'
    mysys.sens.set_status(True)
    mysys.flush()
    assert mysys.act.program == mysys.prog
    assert mysys.act.program_status == {mysys.defprog: LOW, mysys.prog: HIGH}
    assert len(mysys.act.program_stack) == 2

    assert mysys.sens.status == True
    assert mysys.act.get_program_status(mysys.prog) == HIGH
    assert mysys.prog.active == True
    assert mysys.act.status == HIGH
    assert mysys.act.program.name == 'prog'

    mysys.sens.set_status(False)
    mysys.flush()
    assert mysys.act.program.name == 'dp_act'
    assert mysys.act.program == mysys.defprog
    assert len(mysys.act.program_stack) == 1
    assert mysys.act.program_status == {mysys.defprog: LOW}


def test_back_and_forth_on_activate(mysys):
    assert mysys.act.status == LOW
    mysys.sens.set_status(True)
    mysys.flush()
    assert mysys.act.status == HIGH
    mysys.sens.set_status(False)
    mysys.flush()
    assert mysys.act.status == LOW
    logger.debug('STARTING')
    for i in range(AMOUNT):
        mysys.sens.set_status(True)
        mysys.sens.set_status(False)
    mysys.flush()
    assert mysys.sens.status == False
    assert mysys.act.status == LOW


def test_back_and_forth_on_activate_w_safety_delay(mysys):
    mysys.act.safety_delay = 5*DELAYTIME
    mysys.act.debug = True
    assert mysys.act.status == LOW
    mysys.sens.set_status(True)
    mysys.sens.set_status(False)
    mysys.sens.set_status(True)
    mysys.flush()
    assert mysys.act.status == LOW
    mysys.act._last_changed = time.time()
    for i in range(AMOUNT):
        mysys.sens.set_status(True)
        mysys.sens.set_status(False)
    mysys.sens.set_status(True)
    mysys.flush()
    assert mysys.worker_thread.queue.queue == deque([])
    assert mysys.worker_thread.queue.unfinished_tasks == 0
    assert mysys.act.status == LOW


def test_back_and_forth_on_update(mysys):
    assert mysys.act.status == LOW
    mysys.prog.active_condition = Value(True)
    mysys.prog.update_condition = Value(True)
    mysys.prog.on_update = SetStatus(mysys.act, mysys.sens)
    mysys.sens.set_status(True)
    mysys.sens.set_status(False)
    mysys.flush()
    assert mysys.act.status == 0.0

    mysys.sens.set_status(True)
    mysys.flush()
    assert mysys.act.status == 1.0
    mysys.sens.set_status(False)
    mysys.flush()
    assert mysys.act.status == 0.0
    # what about if we quickly turn true&false?
    import random
    for i in range(AMOUNT):
        mysys.sens.set_status(True)
        # print i, mysys.worker_thread.queue.queue
        if random.random() < 0.01:
            time.sleep(0.01)
        mysys.sens.set_status(False)
        # print i, mysys.worker_thread.queue.queue
        # if random.random() < 0.02:
        #    time.sleep(0.01)
    mysys.flush()
    assert mysys.act.status == 0.0
    #assert False


def test_change_delay_with_two_progs(mysys):
    mysys.sens.set_status(True)
    mysys.flush()
    assert mysys.act.status == HIGH
    mysys.act.change_delay = DELAYTIME
    mysys.act.change_mode = 'both'
    mysys.sens.set_status(False)
    mysys.flush()
    assert mysys.act.status == HIGH
    assert mysys.act.changing
    mysys.s1.set_status(True)
    mysys.flush()
    assert mysys.act.status == HIGH
    assert not mysys.act.changing
    mysys.act._last_changed = time.time()
    time.sleep(SLEEPTIME)
    assert mysys.act.status == HIGH
    assert not mysys.act.changing


def test_change_delay_rising(mysys):
    assert mysys.act.safety_mode == 'rising'
    mysys.act.change_delay = DELAYTIME
    mysys.sens.set_status(True)
    mysys.flush()
    assert mysys.act.status == LOW
    assert mysys.act.changing == True
    time.sleep(SLEEPTIME)
    assert mysys.act.changing == False
    assert mysys.act.status == HIGH
    mysys.act._last_changed = time.time()
    mysys.sens.set_status(False)
    mysys.flush()
    assert mysys.act.status == LOW


def test_change_delay_both(mysys):
    mysys.act.change_delay = DELAYTIME
    mysys.act.change_mode = 'both'
    mysys.sens.set_status(True)
    mysys.flush()
    assert mysys.act.status == LOW
    assert mysys.act.changing == True
    time.sleep(SLEEPTIME)
    assert mysys.act.changing == False
    assert mysys.act.status == HIGH
    mysys.act._last_changed = time.time()
    mysys.sens.set_status(False)
    mysys.flush()
    assert mysys.act.status == HIGH
    assert mysys.act.changing == True
    time.sleep(SLEEPTIME)
    assert mysys.act.changing == False
    assert mysys.act.status == LOW


def test_change_delay_falling(mysys):
    mysys.act.change_delay = DELAYTIME
    mysys.act.change_mode = 'falling'
    mysys.act._last_changed = time.time()
    mysys.sens.set_status(True)
    mysys.flush()
    assert mysys.act.status == HIGH
    mysys.sens.set_status(False)
    mysys.flush()
    assert mysys.act.status == HIGH
    assert mysys.act.changing == True
    time.sleep(SLEEPTIME)
    assert mysys.act.changing == False
    assert mysys.act.status == LOW


def test_change_delay_falling_issue23(mysys):
    mysys.act.change_delay = DELAYTIME  # = 3
    mysys.act.change_mode = 'falling'
    mysys.act._last_changed = time.time()
    mysys.sens.set_status(True)  # rising
    mysys.flush()
    mysys.sens.set_status(False)
    mysys.flush()
    mysys.sens.set_status(True)
    mysys.flush()
    assert mysys.act.status == HIGH
    time.sleep(0.9 * DELAYTIME)
    assert mysys.act.status == HIGH
    mysys.sens.set_status(False)
    time.sleep(0.9 * DELAYTIME)
    assert mysys.act.status == HIGH
    time.sleep(0.4 * DELAYTIME)  # must wait full change_delay after set_status(False)!
    assert mysys.act.status == LOW


def test_change_delay_falling_issue23_2(mysys):
    mysys.act.change_delay = DELAYTIME  # = 3
    mysys.act.change_mode = 'falling'
    mysys.act._last_changed = time.time()
    mysys.sens.set_status(True)  # rising
    mysys.flush()
    mysys.sens.set_status(False)
    mysys.flush()
    mysys.sens.set_status(True)
    mysys.flush()
    mysys.sens.set_status(False)
    assert mysys.act.status == HIGH
    time.sleep(0.9 * DELAYTIME)
    assert mysys.act.status == HIGH
    mysys.sens.set_status(False)
    mysys.flush()
    time.sleep(0.3 * DELAYTIME)
    assert mysys.act.status == LOW


def test_safety_delay_with_safety_mode_rising(mysys):
    mysys.act.safety_delay = DELAYTIME
    mysys.act._last_changed = 0.
    assert mysys.act.safety_mode == 'rising'
    assert mysys.sens.status == False
    assert mysys.sens.change_delay == 0.0

    mysys.sens.set_status(True)
    mysys.flush()
    assert mysys.act.changing == False
    assert mysys.act.status == HIGH
    mysys.sens.set_status(False)
    mysys.flush()
    assert mysys.act.changing == False
    assert mysys.act.status == LOW

    mysys.sens.set_status(True)
    mysys.flush()
    assert mysys.act.changing == True
    assert mysys.act.status == LOW
    time.sleep(SLEEPTIME)
    assert mysys.act.changing == False
    assert mysys.act.status == HIGH

    # reason for these: this caused at one point hanging
    mysys.sens.set_status(False)
    mysys.sens.set_status(True)
    mysys.sens.set_status(False)
    mysys.sens.set_status(True)
    mysys.sens.set_status(False)
    mysys.sens.set_status(True)


def test_safety_delay_with_safety_mode_both(mysys):
    mysys.act.safety_delay = DELAYTIME
    mysys.act._last_changed = 0.
    mysys.act.safety_mode = 'both'
    assert mysys.sens.status == False
    assert mysys.sens.change_delay == 0.0

    mysys.sens.set_status(True)
    mysys.flush()
    assert mysys.act.changing == False
    assert mysys.act.status == HIGH

    mysys.act._last_changed = time.time()
    # set time so that there will not accumulate time difference that might fail test sometimes

    mysys.sens.set_status(False)
    mysys.flush()
    assert mysys.act.changing == True
    assert mysys.act.status == HIGH
    time.sleep(SLEEPTIME)
    assert mysys.act.changing == False
    assert mysys.act.status == LOW

    mysys.act._last_changed = time.time()

    mysys.sens.set_status(True)
    mysys.flush()
    assert mysys.sens._status == True
    assert mysys.act.status == LOW
    assert mysys.act.changing == True
    time.sleep(SLEEPTIME)
    assert mysys.act.changing == False
    assert mysys.act.status == HIGH


def test_safety_delay_with_safety_mode_falling(mysys):
    mysys.act.safety_delay = DELAYTIME
    mysys.act._last_changed = 0.
    mysys.act.safety_mode = 'falling'
    assert mysys.sens.status == False
    assert mysys.sens.change_delay == 0.0

    mysys.sens.set_status(True)
    # mysys.flush()
    # assert mysys.sens.changing  # could have changed already, in theory
    #assert mysys.sens._queued_job
    mysys.flush()
    assert mysys.act.changing == False
    assert mysys.act.status == HIGH
    mysys.sens.set_status(False)
    mysys.flush()
    assert mysys.act.changing == True
    assert mysys.act.status == HIGH
    time.sleep(SLEEPTIME)
    assert mysys.act.changing == False
    assert mysys.act.status == LOW

    mysys.sens.set_status(True)
    mysys.flush()
    assert mysys.act.changing == False
    assert mysys.act.status == HIGH


def test_statuschange(mysys, caplog):
    for a in [mysys.s1, mysys.slavebinact]:
        a.change_delay = 5
        a.set_status(True)
        mysys.flush()
        assert a.changing
        mysys.flush()
        assert a._status == False
        assert a._timed_action
        assert not '%s:INFO:Cancelling previous safety/change_delay action. Now changing to False' % a.name in caplog.text()
        a.set_status(False)
        mysys.flush()
        assert not a.changing
        assert not a._timed_action
        assert '%s:INFO:Cancelling previous safety/change_delay action. Now changing to False' % a.name in caplog.text()
        assert a._status == False
        assert not a.changing
        assert not a._timed_action


def test_sensor_resetdelay(sysloader):
    class mysys(System):
        s = UserIntSensor(default=2, reset_delay=0.1)
    s = sysloader.new_system(mysys)
    s.s.status = 1
    s.flush()
    assert s.s.status == 1
    time.sleep(0.2)
    assert s.s.status == 2


class MySensor(AbstractSensor):
    setup_called = CBool(False)

    def setup(self):
        self.setup_called = True


class MyActuator(AbstractActuator):
    setup_called = CBool(False)

    def setup(self):
        self.setup_called = True


def test_statusobject_setup(sysloader):
    class newsys(System):
        a = MySensor()
        b = MyActuator()
    s = sysloader.new_system(newsys)
    assert s.a.setup_called
    assert s.b.setup_called


class TestActuator(FloatActuator):
    change_count = CInt(0)

    def _status_changed(self):
        self.change_count += 1


def test_customactuator_statuschange(sysloader):
    class sys(System):
        testact = TestActuator(default=0.0)
        sens = UserFloatSensor(default=0.0)
        prog = Program(on_update=SetStatus(testact, sens))
    s = sysloader.new_system(sys)
    assert s.testact.change_count == 0
    s.sens.status = 1.0
    s.flush()
    assert s.testact.change_count == 1
    s.sens.status = 1.0
    s.flush()
    assert s.testact.change_count == 1
    s.sens.status = 2.0
    s.flush()
    assert s.testact.change_count == 2
    s.sens.status = 1.0
    s.flush()
    assert s.testact.change_count == 3


def test_on_activate_bug(sysloader):
    called = []

    def func(caller):
        called.append(1)

    class sys(System):
        after_reset = Program(
            update_condition=Value(False),
            on_update=Func(func),
        )

    s = sysloader.new_system(sys)
    assert not called

from traits.api import Int


def test_set_status_wrong_type(sysloader, caplog):
    caplog.error_ok = True

    class mysens(AbstractSensor):
        _status = Int

    class sys(System):
        s = mysens(default=0)
    s = sysloader.new_system(sys)
    s.s.status = 1.1
    s.flush()
    assert s.s.status == 0



# (c) 2015-2017 Tuomas Airaksinen
# 
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.

"""
The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/1.9/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  url(r'^$', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  url(r'^$', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.conf.urls import url, include
    2. Add a URL to urlpatterns:  url(r'^blog/', include('blog.urls'))
"""

from django.conf.urls import url

from . import views

urlpatterns = [
    url(r'^login$', views.login, name='login'),
    url(r'^$', views.main, name='main'),
    url(r'^logout$', views.logout, name='logout'),
    url('^custom/(\w*)$', views.custom, name='custom'),
    url('^uml/system.svg$', views.puml_svg, name='puml_svg'),
    url('^uml/system.puml$', views.puml_raw, name='puml_raw'),
    url('^uml$', views.plantuml, name='plantuml'),

    url('^system$', views.system_view, name='system'),
    url("^threads$", views.threads, name='threads'),

    url('^tag/([\w:]*)', views.single_tag, name='single_tag'),

    url('^tag$', views.tags_view, name='tags_view'),
    url('^user_editable_view$', views.user_defined_view, name='user_editable_view'),
    url('^user_defined_view$', views.tags_view, name='user_defined_view'),
    url('^only_groups$', views.tag_view_only_groups, name='tag_view_only_groups'),
    url('^info_panel/(\w*)', views.info_panel, name='info_panel'),
    url('^toggle_sensor/(\w*)$', views.toggle_sensor, name='toggle_sensor'),
    url(r'^toggle/(\w*)$', views.toggle_value, name='toggle_value'),
    url(r'^set/(\w*)/(\w*)$', views.set_value, name='set_value'),
    url(r'^set_ready/(\w*)/(\w*)', views.set_ready, name='set_ready'),
    url(r'^reload_service/(\d*)$', views.reload_service, name='reload_service'),
    url(r'^cancel_thread/(\d*)$', views.cancel_thread, name='cancel_thread'),
    url(r'^edit/(\w*)$', views.edit, name='edit'),
    url(r'^continue_edit/(\w*)/(\w*)$', views.continue_edit, name='continue_edit'),
    url(r'^new/(\w*)$', views.new, name='new'),
    url('^console$', views.console, name='console'),

    url('^set_status$', views.set_status, name='set_status'),
    url('^.*', views.notfound, name='notfound'),
]


# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of automate-arduino.
#
# automate-arduino is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# automate-arduino is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with automate-arduino.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals
import os
import threading
import logging
from collections import namedtuple

from builtins import range

from automate import Lock
from traits.api import HasTraits, Any, Dict, CList, Str, Int, List
from automate.service import AbstractSystemService

logger = logging.getLogger('automate.arduino_service')

PinTuple = namedtuple('PinTuple', ['type', 'pin'])

def patch_pyfirmata():
    """ Patch Pin class in pyfirmata to have Traits. Particularly, we need notification
        for value changes in Pins. """

    import pyfirmata
    if getattr(pyfirmata, 'patched', False):
        return

    PinOld = pyfirmata.Pin

    class Pin(PinOld, HasTraits):
        mode = property(PinOld._get_mode, PinOld._set_mode)

        def __init__(self, *args, **kwargs):
            HasTraits.__init__(self)
            self.add_trait("value", Any)
            PinOld.__init__(self, *args, **kwargs)

    import pyfirmata.pyfirmata

    pyfirmata.Pin = Pin
    pyfirmata.pyfirmata.Pin = Pin

    from pyfirmata.util import Iterator as OldIterator

    class FixedPyFirmataIterator(OldIterator):

        def run(iter_self):
            try:
                super(FixedPyFirmataIterator, iter_self).run()
            except Exception as e:
                logger.error('Exception %s occurred in Pyfirmata iterator, quitting now', e)
                logger.error('threads: %s', threading.enumerate())

    pyfirmata.util.Iterator.Fixed = FixedPyFirmataIterator
    pyfirmata.patched = True


class ArduinoService(AbstractSystemService):

    """
        Service that provides interface to Arduino devices via
        `pyFirmata library <https://github.com/tino/pyFirmata>`_.
    """

    #: Arduino devices to use, as a list
    arduino_devs = CList(Str, ["/dev/ttyUSB0"])

    #: Arduino device board types, as a list of strings. Choices are defined by pyFirmata board
    #: class names, i.e. allowed values are "Arduino", "ArduinoMega", "ArduinoDue".
    arduino_dev_types = CList(Str, ["Arduino"])

    #: Arduino device sampling rates, as a list (in milliseconds).
    arduino_dev_sampling = CList(Int, [500])

    _sens_analog = Dict
    _sens_digital = Dict
    _act_digital = Dict
    _boards = List
    _locks = List
    _iterator_thread = Any

    class FileNotReadableError(Exception):
        pass

    def setup(self):
        self.logger.debug("Initializing Arduino subsystem")
        try:
            import pyfirmata
        except ImportError:
            self.logger.error("Please install pyfirmata if you want to use Arduino interface")
            return

        patch_pyfirmata()
        from pyfirmata.util import Iterator, to_two_bytes

        # Initialize configured self.boards
        ard_devs = self.arduino_devs
        ard_types = self.arduino_dev_types
        samplerates = self.arduino_dev_sampling
        if not (len(ard_devs) == len(ard_types) == len(samplerates)):
            raise RuntimeError('Arduino configuration invalid!')

        for i in range(len(ard_devs)):
            try:
                if not os.access(ard_devs[i], os.R_OK):
                    raise self.FileNotReadableError
                cls = getattr(pyfirmata, ard_types[i])
                board = cls(ard_devs[i])
                board.send_sysex(pyfirmata.SAMPLING_INTERVAL, to_two_bytes(samplerates[i]))
                self._iterator_thread = it = Iterator.Fixed(board)
                it.daemon = True
                it.name = "PyFirmata thread for {dev}".format(dev=ard_devs[i])
                board._iter = it
                it.start()
                self._boards.append(board)
                self.is_mocked = False
            except (self.FileNotReadableError, OSError) as e:
                if isinstance(e, self.FileNotReadableError) or e.errno == os.errno.ENOENT:
                    self.logger.warning('Your arduino device %s is not available. Arduino will be mocked.', ard_devs[i])
                    self._boards.append(None)
                    self.is_mocked = True
                else:
                    raise
            self._locks.append(Lock())

    def cleanup(self):
        self.logger.debug("Cleaning up Arduino subsystem. ")
        while self._boards:
            board = self._boards.pop()
            if board:
                board.exit()
        if self._iterator_thread and self._iterator_thread.is_alive():
            self._iterator_thread.board = None
            self._iterator_thread.join()
            self._iterator_thread = None

    def reload(self):
        digital_sensors = list(self._sens_digital.items())
        analog_sensors = list(self._sens_analog.items())
        digital_actuators = list(self._act_digital.items())

        for (dev, pin_nr), (_type, pin) in digital_actuators:
            self.cleanup_digital_actuator(dev, pin_nr)

        for (dev, pin_nr), (sens, pin) in digital_sensors:
            self.unsubscribe_digital(dev, pin_nr)
        for (dev, pin_nr), (sens, pin) in analog_sensors:
            self.unsubscribe_analog(dev, pin_nr)
        super(ArduinoService, self).reload()
        # Restore subscriptions
        for (dev, pin_nr), (sens, pin) in digital_sensors:
            self.subscribe_digital(dev, pin_nr, sens)
        for (dev, pin_nr), (sens, pin) in analog_sensors:
            self.subscribe_analog(dev, pin_nr, sens)

        for (dev, pin_nr), (_type, pin) in digital_actuators:
            setup_func = {'p': self.setup_pwm, 'o': self.setup_digital}.get(_type)
            # TODO: servo reload not implemented
            if setup_func:
                setup_func(dev, pin_nr)
            else:
                logger.error('Reloading not implemented for type %s (pin %d)', _type, pin_nr)
        self.logger.info('Arduino pins are now set up!')

    def setup_digital(self, dev, pin_nr):
        if not self._boards[dev]:
            self._act_digital[(dev, pin_nr)] = PinTuple('o', None)
            return
        with self._locks[dev]:
            pin = self._boards[dev].get_pin("d:{pin}:o".format(pin=pin_nr))
            self._act_digital[(dev, pin_nr)] = PinTuple('o', pin)

    def setup_pwm(self, dev, pin_nr):
        if not self._boards[dev]:
            self._act_digital[(dev, pin_nr)] = PinTuple('p', None)
            return
        with self._locks[dev]:
            pin = self._boards[dev].get_pin("d:{pin}:p".format(pin=pin_nr))
            self._act_digital[(dev, pin_nr)] = PinTuple('p', pin)

    def setup_servo(self, dev, pin_nr, min_pulse, max_pulse, angle):
        if not self._boards[dev]:
            self._act_digital[(dev, pin_nr)] = PinTuple('s', None)
            return
        with self._locks[dev]:
            pin = self._boards[dev].get_pin("d:{pin}:s".format(pin=pin_nr))
            self._act_digital[(dev, pin_nr)] = PinTuple('s', pin)
            self._boards[dev].servo_config(pin_nr, min_pulse, max_pulse, angle)

    def change_digital(self, dev, pin_nr, value):
        """ Change digital Pin value (boolean). Also PWM supported(float)"""
        if not self._boards[dev]:
            return
        with self._locks[dev]:
            self._act_digital[(dev, pin_nr)].pin.write(value)

    # Functions for input signals
    def handle_analog(self, obj, name, old, new):
        dev = obj.__dev_id
        pin = obj.pin_number
        if not self._boards[dev]:
            return
        self._sens_analog[(dev, pin)][0].set_status(new)

    def handle_digital(self, obj, name, old, new):
        dev = obj.__dev_id
        pin = obj.pin_number
        if not self._boards[dev]:
            return
        self._sens_digital[(dev, pin)][0].set_status(new)

    def subscribe_analog(self, dev, pin_nr, sens):
        if not self._boards[dev]:
            return
        with self._locks[dev]:
            pin = self._boards[dev].get_pin("a:{pin}:i".format(pin=pin_nr))
            pin.__dev_id = dev
            self._sens_analog[(dev, pin_nr)] = (sens, pin)
            s = pin.read()
        if s is not None:
            sens.set_status(s)
        pin.on_trait_change(self.handle_analog, "value")

    def cleanup_digital_actuator(self, dev, pin_nr):
        pin = self._act_digital.pop((dev, pin_nr), None)

    def unsubscribe_digital(self, dev, pin_nr):
        pin = self._sens_digital.pop((dev, pin_nr), None)
        if pin:
            pin[1].remove_trait('value')

    def unsubscribe_analog(self, dev, pin_nr):
        pin = self._sens_analog.pop((dev, pin_nr), None)
        if pin:
            pin[1].remove_trait('value')

    def subscribe_digital(self, dev, pin_nr, sens):
        if not self._boards[dev]:
            return
        with self._locks[dev]:
            pin = self._boards[dev].get_pin("d:{pin}:i".format(pin=pin_nr))
            pin.__dev_id = dev
            self._sens_digital[(dev, pin_nr)] = (sens, pin)
            s = pin.read()
        if s is not None:
            sens.set_status(s)
        pin.on_trait_change(self.handle_digital, "value")



# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals
from past.builtins import basestring
from collections import defaultdict

from future import standard_library
standard_library.install_aliases()
from builtins import input
import threading
import operator
import sys
import os
import logging
import pickle
import pkg_resources
import argparse

from traits.api import (CStr, Instance, CBool, CList, Property, CInt, CUnicode, Event, CSet, Str, cached_property,
                        on_trait_change)

from automate.common import (SystemBase, ExitException, has_baseclass, Object)
from automate.namespace import Namespace
from automate.service import AbstractService, AbstractUserService, AbstractSystemService
from automate.statusobject import AbstractSensor, AbstractActuator
from automate.systemobject import SystemObject
from automate.worker import StatusWorkerThread
from automate.callable import AbstractCallable

import sys

if sys.version_info >= (3, 0):
    TimerClass = threading.Timer
else:
    TimerClass = threading._Timer

def get_autoload_services():
    import automate.services
    return (i for i in list(automate.services.__dict__.values()) if has_baseclass(i, AbstractService) and i.autoload)


def get_service_by_name(name):
    import automate.services
    return getattr(automate.services, name)


class System(SystemBase):
    #: Name of the system (shown in WEB UI for example)
    name = CStr

    #: Allow referencing objects by their names in Callables. If disabled, you can still refer to objects by names
    #: by Object('name')
    allow_name_referencing = CBool(True)

    #: Filename to where to dump the system state
    filename = Str

    # LOGGING
    ###########

    #: Name of the file where logs are stored
    logfile = CUnicode

    #: Log level for the handler that writes to stdout
    print_level = CInt(logging.INFO, transient=True)

    @on_trait_change('print_level', post_init=True)
    def print_level_changed(self, new):
        self.print_handler.setLevel(new)

    #: Reference to logger instance (read-only)
    logger = Instance(logging.Logger)

    #: Instance to the log handler that writes to stdout
    log_handler = Instance(logging.Handler)

    #: Format string of the log handler that writes to stdout
    log_format = Str('%(asctime)s %(log_color)s%(name)s%(reset)s %(message)s')

    #: Instance to the log handler that writes to logfile (read-only)
    print_handler = Instance(logging.Handler)

    #: Format string of the log handler that writes to logfile
    logfile_format = Str('%(process)d:%(threadName)s:%(name)s:%(asctime)s:%(levelname)s:%(message)s')

    #: Log level of the handler that writes to logfile
    log_level = CInt(logging.DEBUG, transient=True)

    @on_trait_change('log_level', post_init=True)
    def log_level_changed(self, new):
        if not self.logfile:
            self.logger.error('No logfile specified')
            return
        self.log_handler.setLevel(new)

    # SERVICES
    ###########

    #: Add here services that you want to be added automatically. This is meant to be re-defined in subclass.
    default_services = CList(trait=Str)

    #: List of services that are loaded in the initialization of the System.
    services = CList(trait=Instance(AbstractService))

    #: List of servicenames that are desired to be avoided (even if normally autoloaded).
    exclude_services = CSet(trait=Str)

    #: Reference to the worker thread (read-only)
    worker_thread = Instance(StatusWorkerThread, transient=True)

    #: System namespace (read-only)
    namespace = Instance(Namespace)

    # Set of all SystemObjects within the system. This is where SystemObjects are ultimately stored
    # in the System initialization. (read-only)
    objects = CSet(trait=SystemObject)

    #: Property giving objects sorted alphabetically (read-only)
    objects_sorted = Property(depends_on='objects')

    @cached_property
    def _get_objects_sorted(self):
        return sorted(list(self.objects), key=operator.attrgetter('_order'))

    #: Read-only property giving all sensors of the system
    sensors = Property(depends_on='objects[]')

    @cached_property
    def _get_sensors(self):
        return {i for i in self.objects_sorted if isinstance(i, AbstractSensor)}

    #: Read-only property giving all actuator of the system
    actuators = Property(depends_on='objects[]')

    @cached_property
    def _get_actuators(self):
        return {i for i in self.objects_sorted if isinstance(i, AbstractActuator)}

    #: Read-only property giving all objects that have program features in use
    programs = Property(depends_on='objects[]')

    @cached_property
    def _get_programs(self):
        from .program import Program, DefaultProgram
        return {i for i in self.objects_sorted if isinstance(i, (Program, DefaultProgram))}

    #: Read-only property giving all :class:`~program.Program` objects
    ordinary_programs = Property(depends_on='programs[]')

    @cached_property
    def _get_ordinary_programs(self):
        from . import program
        return {i for i in self.programs if isinstance(i, program.Program)}

    #: Start worker thread automatically after system is initialized
    worker_autostart = CBool(True)

    #: Trigger which is triggered after initialization is ready (used by Services)
    post_init_trigger = Event

    #: Trigger which is triggered before quiting (used by Services)
    pre_exit_trigger = Event

    #: Read-only property that gives list of all object tags
    all_tags = Property(depends_on='objects.tags[]')

    @cached_property
    def _get_all_tags(self):
        newset = set([])
        for i in self.system.objects:
            for j in i.tags:
                if j:
                    newset.add(j)
        return newset

    #: Enable experimental two-phase queue handling technique (not recommended)
    two_phase_queue = CBool(False)

    @classmethod
    def load_or_create(cls, filename=None, no_input=False, **kwargs):
        """
            Load system from a dump, if dump file exists, or create a new system if it does not exist.
        """
        parser = argparse.ArgumentParser()
        parser.add_argument('--no_input', action='store_true')
        args = parser.parse_args()

        if args.no_input:
            print('Parameter --no_input was given')
            no_input = True

        def savefile_more_recent():
            time_savefile = os.path.getmtime(filename)
            time_program = os.path.getmtime(sys.argv[0])
            return time_savefile > time_program

        def load():
            print('Loading %s' % filename)
            file = open(filename, 'rb')
            state = pickle.load(file)
            file.close()
            system = System(loadstate=state, filename=filename, **kwargs)
            return system

        def create():
            print('Creating new system')
            return cls(filename=filename, **kwargs)

        if filename and os.path.isfile(filename):
            if savefile_more_recent():
                return load()
            else:
                if no_input:
                    print('Program file more recent. Loading that instead.')
                    return create()
                while True:
                    answer = input('Program file more recent. Do you want to load it? (y/n) ')
                    if answer == 'y':
                        return create()
                    elif answer == 'n':
                        return load()
        else:
            return create()

    def save_state(self):
        """
            Save state of the system to a dump file :attr:`System.filename`
        """
        if not self.filename:
            self.logger.error('Filename not specified. Could not save state')
            return
        self.logger.debug('Saving system state to %s', self.filename)
        with open(self.filename, 'wb') as file, self.worker_thread.queue.mutex:
            pickle.dump((list(self.objects)), file, pickle.HIGHEST_PROTOCOL)

    @property
    def cmd_namespace(self):
        """
            A read-only property that gives the namespace of the system for evaluating commands.
        """
        import automate
        ns = dict(list(automate.__dict__.items()) + list(self.namespace.items()))
        return ns

    def __getattr__(self, item):
        if self.namespace and item in self.namespace:
            return self.namespace[item]
        raise AttributeError

    def get_unique_name(self, obj, name='', name_from_system=''):
        """
            Give unique name for an Sensor/Program/Actuator object
        """
        ns = self.namespace
        newname = name
        if not newname:
            newname = name_from_system

        if not newname:
            newname = u"Nameless_" + obj.__class__.__name__

        if not newname in ns:
            return newname

        counter = 0
        while True:
            newname1 = u"%s_%.2d" % (newname, counter)
            if not newname1 in ns:
                return newname1
            counter += 1

    @property
    def services_by_name(self):
        """
            A property that gives a dictionary that contains services as values and their names as keys.
        """
        srvs = defaultdict(list)
        for i in self.services:
            srvs[i.__class__.__name__].append(i)
        return srvs

    @property
    def service_names(self):
        """
            A property that gives the names of services as a list
        """
        return set(self.services_by_name.keys())

    def flush(self):
        """
            Flush the worker queue. Usefull in unit tests.
        """
        self.worker_thread.flush()

    def name_to_system_object(self, name):
        """
            Give SystemObject instance corresponding to the name
        """
        if isinstance(name, basestring):
            if self.allow_name_referencing:
                name = name
            else:
                raise NameError('System.allow_name_referencing is set to False, cannot convert string to name')
        elif isinstance(name, Object):
            name = str(name)
        return self.namespace.get(name, None)

    def eval_in_system_namespace(self, exec_str):
        """
            Get Callable for specified string (for GUI-based editing)
        """
        ns = self.cmd_namespace
        try:
            return eval(exec_str, ns)
        except Exception as e:
            self.logger.warning('Could not execute %s, gave error %s', exec_str, e)
            return None

    def register_service_functions(self, *funcs):
        """
            Register function in the system namespace. Called by Services.
        """
        for func in funcs:
            self.namespace[func.__name__] = func

    def register_service(self, service):
        """
            Register service into the system. Called by Services.
        """
        if service not in self.services:
            self.services.append(service)

    def request_service(self, type, id=0):
        """
            Used by Sensors/Actuators/other services that need to use other services for their
            operations.
        """
        srvs = self.services_by_name.get(type)
        if not srvs:
            return

        ser = srvs[id]

        if not ser.system:
            ser.setup_system(self)
        return ser

    def cleanup(self):
        """
            Clean up before quitting
        """

        self.pre_exit_trigger = True

        self.logger.info("Shutting down %s, please wait a moment.", self.name)
        for t in threading.enumerate():
            if isinstance(t, TimerClass):
                t.cancel()
        self.logger.debug('Timers cancelled')

        for i in self.objects:
            i.cleanup()
            del i

        self.logger.debug('Sensors etc cleanups done')

        for ser in (i for i in self.services if isinstance(i, AbstractUserService)):
            ser.cleanup_system()
        self.logger.debug('User services cleaned up')

        self.worker_thread.stop()
        self.logger.debug('Worker thread really stopped')

        for ser in (i for i in self.services if isinstance(i, AbstractSystemService)):
            ser.cleanup_system()
        self.logger.debug('System services cleaned up')

    def cmd_exec(self, cmd):
        """
            Execute commands in automate namespace
        """

        if not cmd:
            return
        ns = self.cmd_namespace
        import copy
        rval = True
        nscopy = copy.copy(ns)
        try:
            r = eval(cmd, ns)
            if isinstance(r, SystemObject) and not r.system:
                r.setup_system(self)
            if callable(r):
                r = r()
                cmd += "()"
            self.logger.info("Eval: %s", cmd)
            self.logger.info("Result: %s", r)
        except SyntaxError:
            r = {}
            try:
                exec (cmd, ns)
                self.logger.info("Exec: %s", cmd)
            except ExitException:
                raise
            except Exception as e:
                self.logger.info("Failed to exec cmd %s: %s.", cmd, e)
                rval = False
            for key, value in list(ns.items()):
                if key not in nscopy or not value is nscopy[key]:
                    if key in self.namespace:
                        del self.namespace[key]
                    self.namespace[key] = value
                    r[key] = value
            self.logger.info("Set items in namespace: %s", r)
        except ExitException:
            raise
        except Exception as e:
            self.logger.info("Failed to eval cmd %s: %s", cmd, e)
            return False

        return rval

    def __init__(self, loadstate=None, **traits):
        super(System, self).__init__(**traits)
        if not self.name:
            self.name = os.path.split(sys.argv[0])[-1].replace('.py', '')

        self.worker_thread = StatusWorkerThread(name="Status worker thread")

        self._initialize_logging()
        self.logger.info('Initializing services')
        self._initialize_services()
        self.logger.info('Initializing namespace')
        self._initialize_namespace(loadstate)
        self.logger.info('Initialize user services')
        self._setup_user_services()

        if self.worker_autostart:
            self.logger.info('Starting worker thread')
            self.worker_thread.start()

        self.post_init_trigger = True

    def _initialize_logging(self):
        self.logger = logging.getLogger()
        if len(self.logger.handlers) > 0:
            self.logger.warning('Logging has been setup already')
            return

        self.logger.setLevel(logging.DEBUG)

        # Why is threre StreamHandler when running gpiotest.py from raspi2?
        # Reason: RPIO uses logging.error etc directly, which adds handler
        # to root. .

        formatter = logging.Formatter(fmt=self.logfile_format)

        if self.logfile:
            self.log_handler = log_handler = logging.FileHandler(self.logfile)
            log_handler.setLevel(self.log_level)
            log_handler.setFormatter(formatter)
            self.logger.addHandler(log_handler)

        self.print_handler = streamhandler = logging.StreamHandler()
        streamhandler.setLevel(self.print_level)

        from colorlog import ColoredFormatter, default_log_colors
        colors = default_log_colors.copy()
        colors['DEBUG'] = 'purple'

        streamhandler.setFormatter(ColoredFormatter(self.log_format, datefmt='%H:%M:%S', log_colors=colors))

        self.logger.addHandler(streamhandler)
        self.logger = logging.getLogger(self.name)
        self.logger.info('Logging setup ready')

    def _initialize_namespace(self, loadstate=None):
        self.namespace = Namespace(system=self)
        self.namespace.set_system(loadstate)

        self.logger.info('Setup loggers per object')
        for k, v in list(self.namespace.items()):
            if isinstance(v, SystemObject):
                ctype = v.__class__.__name__
                v.logger = self.logger.getChild('%s.%s' % (ctype, k))

    def _initialize_services(self):
        # Add default_services, if not already
        for servname in self.default_services:
            if servname not in self.service_names | self.exclude_services:
                self.services.append(get_service_by_name(servname)())

        # Add autorun services if not already
        for servclass in get_autoload_services():
            if servclass.__name__ not in self.service_names | self.exclude_services:
                self.services.append(servclass())

    def _setup_user_services(self):
        for ser in (i for i in self.services if isinstance(i, AbstractUserService)):
            self.logger.info('...%s', ser.__class__.__name__)
            ser.setup_system(self)


# Load extensions

from . import services, sensors, actuators, callables
print('Loading extensions')
for entry_point in pkg_resources.iter_entry_points('automate.extension'):
    print('Trying to load extension %s' % entry_point)
    try:
        ext_classes = entry_point.load(require=False)
    except ImportError:
        print('Loading extension %s failed. Perhaps missing requirements? Skipping.' % entry_point)
        continue
    for ext_class in ext_classes:
        print('... %s' % ext_class.__name__)
        if issubclass(ext_class, AbstractService):
            setattr(services, ext_class.__name__, ext_class)
        elif issubclass(ext_class, AbstractSensor):
            setattr(sensors, ext_class.__name__, ext_class)
        elif issubclass(ext_class, AbstractActuator):
            setattr(actuators, ext_class.__name__, ext_class)
        elif issubclass(ext_class, AbstractCallable):
            setattr(callables, ext_class.__name__, ext_class)



# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

from __future__ import unicode_literals
from threading import Timer

from traits.api import Unicode, Any, CBool, CFloat

from automate.service import AbstractUserService

__all__ = ['StatusSaverService']


class StatusSaverService(AbstractUserService):

    """
        Service which is responsible for scheduling dumping system into file periodically.
    """

    autoload = True

    #: Dump saving interval, in seconds. Default 30 minutes.
    dump_interval = CFloat(30 * 60)

    _exit = CBool(False)
    _timer = Any(transient=True)

    def setup(self):
        if self.system.filename:
            self.system.on_trait_change(self.exit_save, "pre_exit_trigger")
            if self.dump_interval:
                self.save_system_periodically()

    def save_system_periodically(self):
        self.logger.debug('Saving system state')
        self.system.save_state()
        self._timer = Timer(self.dump_interval, self.save_system_periodically)
        self._timer.start()

    def exit_save(self):
        self.system.save_state()
        self._exit = True

    def cleanup(self):
        if self._timer and self._timer.is_alive():
            self._timer.cancel()



from __future__ import unicode_literals
# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

from automate import *
import pytest, mock
from datetime import datetime


def unix_time(dt):
    epoch = datetime.utcfromtimestamp(0)
    delta = dt - epoch
    return delta.total_seconds()


@pytest.mark.parametrize('_timer_on, _timer_off, now, status', [
    ('0 10 * * *', '0 11 * * *', datetime(2015, 1, 1, 10, 30), True),
    ('0 10 * * *', '0 11 * * *', datetime(2015, 1, 1, 11, 30), False),
    ('0 10 * * *', '0 11 * * *', datetime(2015, 1, 1, 9, 30), False),
    ('0 10 * * *;0 12 * * *', '0 11 * * *; 0 13 * * *', datetime(2015, 1, 1, 10, 30), True),
    ('0 10 * * *;0 12 * * *', '0 11 * * *; 0 13 * * *', datetime(2015, 1, 1, 11, 30), False),
    ('0 10 * * *;0 12 * * *', '0 11 * * *; 0 13 * * *', datetime(2015, 1, 1, 12, 30), True),
    ('0 10 * * *;0 12 * * *', '0 11 * * *; 0 13 * * *', datetime(2015, 1, 1, 10, 00), True),
    ('0 10 * * *;0 12 * * *', '0 11 * * *; 0 13 * * *', datetime(2015, 1, 1, 11, 00), False),
    ('0 10 * * *;0 12 * * *;0 15 * * *', '0 11 * * *; 0 13 * * *', datetime(2015, 1, 1, 15, 00), True),
    ('0 10 * * *;0 12 * * *;0 15 * * *', '0 11 * * *; 0 13 * * *', datetime(2015, 1, 1, 9, 30), True),
    ('0 10 * * *;0 12 * * *;0 15 * * *', '0 11 * * *; 0 13 * * *', datetime(2015, 1, 1, 10, 30), True),
    ('0 10 * * *;0 12 * * *;0 15 * * *', '0 11 * * *; 0 13 * * *', datetime(2015, 1, 1, 11, 00), False),
    ('0 10 * * *;0 12 * * *;0 15 * * *', '0 11 * * *; 0 13 * * *', datetime(2015, 1, 1, 13, 00), False),
])
def test_crontimersensor(_timer_on, _timer_off, now, status, sysloader):
    with mock.patch('automate.CronTimerSensor._now', lambda self: now):
        class ms(System):
            t = CronTimerSensor(timer_on=_timer_on, timer_off=_timer_off)

        s = sysloader.new_system(ms)
        assert s.t.status == status



# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

from __future__ import absolute_import
from __future__ import unicode_literals
import logging

from traits.api import cached_property, on_trait_change, CFloat, Instance, CBool, CSet, Property

from automate.common import (LogicStr, Lock, NameOrSensorActuatorBaseTrait,
                             AbstractStatusObject)
from automate.systemobject import SystemObject
from automate.callable import AbstractCallable


class ProgrammableSystemObject(SystemObject):

    """
        System object with standard program features (i.e. conditions & actions).
    """

    simple_view = ['active']
    view = ["name", "priority", "tags", "active_condition_str", "update_condition_str",
            "on_activate_str", "on_deactivate_str", "on_update_str", "triggers_str"] + SystemObject.view

    callables = ['on_activate', 'on_deactivate', 'on_update', 'active_condition', 'update_condition']

    def get_default_callables(self):
        from automate.callables import Value, Empty
        return dict(
            active_condition=Value(True),
            update_condition=Value(True),
            on_activate=Empty(),
            on_update=Empty(),
            on_deactivate=Empty(),
        )

    #: A condition Callable which determines the condition, when the program is activated. Program deactivates, when
    #: condition turns to False. When program is activated, on_activate action is executed. When program
    #: deactivates. on_deactivate is executed.
    active_condition = Instance(AbstractCallable)

    #: An action Callable to be executed when Program actives.
    on_activate = Instance(AbstractCallable)

    #: An action Callable to be executed when Program deactivates.
    on_deactivate = Instance(AbstractCallable)

    #: When program is active, this is the condition Callable that must equal to ``True`` in order to
    #: on_update action to be executed. Whenever a trigger is changed, this condition is checked and
    #: if ``True``, on_update is executed.
    update_condition = Instance(AbstractCallable)

    #: Action Callable to be executed if Program is active and update_condition is ``True``.
    on_update = Instance(AbstractCallable)

    #: When programs sets Actuator status, the actual status of Actuator is determined by a program that has highest
    #: priority. Lower priority programs are stacked and used only if higher priority programs are deactivated.
    priority = CFloat(1)

    #: Is program active? Automatically changed. In UIs you can fake the program active status by changing this.
    #: Normally do not change manually.
    active = CBool(False, transient=True)

    #: Status property is introduced to have interface compability with Status objects.
    #: For plain Programs, status equals to the result of its active condition Callable.
    status = Property(depends_on='active')

    @cached_property
    def _get_status(self):
        return self.active

    _trigger_lock = Instance(Lock, transient=True)

    def __init__(self, *args, **kwargs):
        self._trigger_lock = Lock('triggerlock')
        super(ProgrammableSystemObject, self).__init__(*args, **kwargs)

    def __setstate__(self, *args, **kwargs):
        self._trigger_lock = Lock('triggerlock')
        return super(ProgrammableSystemObject, self).__setstate__(*args, **kwargs)

    #: (read-only property) Set of triggers, that cause this Program conditions to be checked
    #: (and actions to be executed). This data is updated from custom triggers list, conditions and actions.
    actual_triggers = Property(trait=CSet(trait=Instance(AbstractStatusObject)),
                               depends_on='triggers, triggers_items, exclude_triggers, exclude_triggers_items, '
                               'update_condition.triggers, active_condition.triggers, on_update.triggers, '
                               'on_activate.triggers')

    #: (read-only property) Set of targets that this Program might touch. This data is updated
    #: from custom targets list and actions.
    actual_targets = Property(trait=CSet(trait=Instance(AbstractStatusObject)),
                              depends_on='targets, targets_items, on_update.targets, on_activate.targets, '
                                         'on_deactivate.targets')

    #: Custom set of additional triggers, whose status change will trigger this Program conditions/actions
    triggers = CSet(trait=NameOrSensorActuatorBaseTrait)

    #: Triggers in this set do not trigger the program actions/conditions even if they are introduced by
    #: Callables etc.
    exclude_triggers = CSet(trait=NameOrSensorActuatorBaseTrait)

    #: Additional targets. Not usually needed, but if you want to set status for some reason by some custom function,
    #: for example, then you need to use this.
    targets = CSet(trait=NameOrSensorActuatorBaseTrait)

    logger = Instance(logging.Logger, transient=True)

    # Getter and setter used by _str versions of conditions, actions and targets&triggers lists.
    def _str_getter(self, name, trait):
        attr_name = name[:-4]
        attr = getattr(self, attr_name)
        if hasattr(attr, "give_str"):
            return attr.give_str()
        else:
            return str(attr)

    def _str_setter(self, name, value):
        attr_name = name[:-4]
        f = self.system.eval_in_system_namespace(value)
        if f is not None:
            setattr(self, attr_name, f)

    # For UIs, string versions of the Callables
    update_condition_str = Property(depends_on="update_condition, actual_triggers.name, actual_targets.name",
                                    trait=LogicStr, transient=True, fset=_str_setter, fget=_str_getter)
    on_update_str = Property(depends_on="on_update, actual_triggers.name, actual_targets.name", trait=LogicStr,
                             transient=True, fset=_str_setter, fget=_str_getter)
    active_condition_str = Property(depends_on="active_condition, actual_triggers.name, actual_targets.name",
                                    trait=LogicStr, transient=True, fset=_str_setter, fget=_str_getter)
    on_activate_str = Property(depends_on="on_activate, actual_triggers.name, actual_targets.name", trait=LogicStr,
                               transient=True, fset=_str_setter, fget=_str_getter)
    on_deactivate_str = Property(depends_on="on_deactivate, actual_triggers.name, actual_targets.name",
                                 trait=LogicStr, transient=True, fset=_str_setter, fget=_str_getter)

    targets_str = Property(depends_on="targets, targets.name", trait=LogicStr,
                           transient=True, fset=_str_setter, fget=_str_getter)
    triggers_str = Property(depends_on="triggers, triggers.name", trait=LogicStr,
                            transient=True, fset=_str_setter, fget=_str_getter)

    @cached_property
    def _get_actual_triggers(self):
        for c in [self.update_condition, self.active_condition, self.on_update, self.on_activate]:
            c.setup_callable_system(self.system)
        return (self.triggers | self.update_condition.triggers | self.active_condition.triggers
                | self.on_update.triggers | self.on_activate.triggers) - self.exclude_triggers

    @cached_property
    def _get_actual_targets(self):
        for c in [self.on_update, self.on_activate, self.on_deactivate]:
            c.setup_callable_system(self.system)
        return self.targets | self.on_update.targets | self.on_activate.targets | self.on_deactivate.targets

    @on_trait_change('actual_triggers')
    def actual_triggers_changed(self, obj, name, old, new):
        if old is None:
            old = set()
        if old == new:
            return
        self.logger.debug('Actual triggers changed by %s: %s->%s', name, old, new)
        for t in old - new:
            self.logger.debug("Removing trigger %s", t)
            t.on_trait_change(self.trigger_status_changed, "status", remove=True)

        for t in new - old:
            self.logger.debug("Adding trigger %s", t)
            t.on_trait_change(self.trigger_status_changed, "status")

        old_active = self.active
        self.active = bool(self.active_condition.call(self))
        if self.active != old_active:
            self.update_activation(self.active)

    @on_trait_change('actual_targets')
    def actual_targets_changed(self, obj, name,  old, new):
        if old == new:
            return
        if old is None:
            old = set()
        self.logger.debug('Actual targets changed %s->%s', old, new)
        old_active = self.active
        new_active = self.active = bool(self.active_condition.call(self))
        if new_active != old_active:
            self.update_activation(new_active)

        if old_active == new_active == True:
            for i in old - new:
                i.deactivate_program(self)

            for i in new - old:
                i.activate_program(self)

    def update_activation(self, new_active, trigger=None):
        if new_active:
            for t in self.actual_targets:
                t.activate_program(self)
            self.on_activate.setup_callable_system(self.system)
            self.on_activate.call(self, trigger=trigger)
            if bool(self.update_condition.call(self, trigger=trigger)):
                self.on_update.setup_callable_system(self.system)
                self.on_update.cancel(self)
                self.on_update.call(self, trigger=trigger)
        else:
            self.on_deactivate.setup_callable_system(self.system)
            self.on_deactivate.call(self, trigger=trigger)
            self.on_update.cancel(self)
            self.on_activate.cancel(self)
            for t in self.actual_targets:
                t.deactivate_program(self)

    def trigger_status_changed(self, obj, name, old, new):
        self.logger.debug("Trigger status changed from %s %s: %s->%s", obj, name, old, new)
        with self._trigger_lock:
            old_active = self.active
            new_active = self.active = bool(self.active_condition.call(self, trigger=obj))
            if new_active != old_active:
                self.update_activation(new_active, trigger=obj)
            if old_active == new_active == True:
                if bool(self.update_condition.call(self, trigger=obj)):
                    self.on_update.cancel(self)
                    self.on_update.call(self, trigger=obj)
        self.logger.debug("Trigger status changing ready")

    @on_trait_change("active_condition, on_activate, on_deactivate")
    def _update_activation_actions(self, name, new):
        self.logger.debug('Update activation actions %s', name)
        getattr(self, name).setup_callable_system(self.system)

        if name == 'active_condition':
            old_active = self.active
            self.active = bool(self.active_condition.call(self))
            if old_active != self.active:
                self.update_activation(self.active)

        elif name == 'on_activate' and self.active:
            self.on_activate.cancel(self)
            self.on_activate.call(self)

    @on_trait_change('update_condition, on_update')
    def _update_update_actions(self, name, new):
        self.logger.debug('Update update actions %s', name)
        getattr(self, name).setup_callable_system(self.system)

        if self.active and bool(self.update_condition.call(self)):
            self.on_update.cancel(self)
            self.on_update.call(self)

    def setup_system(self, system, *args, **kwargs):
        from .callables import Value
        c = self.get_default_callables()
        c.pop('active_condition')  # We do not want to activate program at this phase.
        c.pop('update_condition')  # nor do we want to run on_update
        self.active_condition = Value(False)
        self.update_condition = Value(False)
        for key, value in list(c.items()):
            setattr(self, key, value)

        super(ProgrammableSystemObject, self).setup_system(system, *args, **kwargs)

    def _priority_changed(self):
        self.logger.debug('Priority changed')
        for t in self.actual_targets:
            t.update_program_stack()


class Program(ProgrammableSystemObject):
    is_program = True
    editable = True


class DefaultProgram(ProgrammableSystemObject):
    editable = True

    def __init__(self, system, name='', **traits):
        self.logger = logging.getLogger('automate.defaultprogram')
        self._passed_arguments = name, traits
        self.setup_system(system)
        self.setup_callables()

    def get_default_callables(self):
        from automate.callables import Value
        callables = super(DefaultProgram, self).get_default_callables()
        callables['active_condition'] = Value(False)
        return callables



#   Copyright (c) 2006-2007 Open Source Applications Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import sys
if sys.version_info < (3, 0):
    from SimpleXMLRPCServer import SimpleXMLRPCDispatcher
else:
    from xmlrpc.server import SimpleXMLRPCDispatcher
import logging

logger = logging.getLogger(__name__)

class WSGIXMLRPCApplication(object):
    """Application to handle requests to the XMLRPC service"""

    def __init__(self, instance=None, methods=None):
        """Create windmill xmlrpc dispatcher"""
        if methods is None:
            methods = []
        try:
            self.dispatcher = SimpleXMLRPCDispatcher(allow_none=True, encoding=None)
        except TypeError:
            # python 2.4
            self.dispatcher = SimpleXMLRPCDispatcher()
        if instance is not None:
            self.dispatcher.register_instance(instance)
        for method in methods:
            self.dispatcher.register_function(method)
        self.dispatcher.register_introspection_functions()

    def handler(self, environ, start_response):
        """XMLRPC service for windmill browser core to communicate with"""

        if environ['REQUEST_METHOD'] == 'POST':
            return self.handle_POST(environ, start_response)
        else:
            start_response("400 Bad request", [('Content-Type','text/plain')])
            return ['']

    def handle_POST(self, environ, start_response):
        """Handles the HTTP POST request.

        Attempts to interpret all HTTP POST requests as XML-RPC calls,
        which are forwarded to the server's _dispatch method for handling.

        Most code taken from SimpleXMLRPCServer with modifications for wsgi and my custom dispatcher.
        """

        try:
            # Get arguments by reading body of request.
            # We read this in chunks to avoid straining
            # socket.read(); around the 10 or 15Mb mark, some platforms
            # begin to have problems (bug #792570).

            length = int(environ['CONTENT_LENGTH'])
            data = environ['wsgi.input'].read(length)

            max_chunk_size = 10*1024*1024
            size_remaining = length

            # In previous versions of SimpleXMLRPCServer, _dispatch
            # could be overridden in this class, instead of in
            # SimpleXMLRPCDispatcher. To maintain backwards compatibility,
            # check to see if a subclass implements _dispatch and
            # using that method if present.
            response = self.dispatcher._marshaled_dispatch(
                    data, getattr(self.dispatcher, '_dispatch', None)
                )
            response += b'\n'
        except Exception as e: # This should only happen if the module is buggy
            # internal error, report as HTTP server error
            logger.error('POST handler crashed due to error %s', e)
            start_response("500 Server error", [('Content-Type', 'text/plain')])
            return []
        else:
            # got a valid XML RPC response
            start_response("200 OK", [('Content-Type','text/xml'), ('Content-Length', str(len(response)),)])
            return [response]

    def __call__(self, environ, start_response):
        return self.handler(environ, start_response)


# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

from __future__ import absolute_import
from __future__ import unicode_literals
import logging
from traits.has_traits import HasStrictTraits, cached_property
from traits.trait_types import Instance, CUnicode, Tuple, Dict, Event, Unicode, Int, CBool
from traits.traits import Property
from automate import SystemBase, TagSet, is_valid_variable_name


class SystemObject(HasStrictTraits):

    """
        Baseclass for Programs, Sensor, Actuators
    """

    #: Names of attributes that accept Callables. If there are custom callables being used, they must be added here.
    #: The purpose of this list is that these Callables will be initialized properly.
    #: :class:`~automate.program.ProgrammableSystemObject` introduces 5 basic callables
    #: (see also :ref:`automate-programs`).
    callables = []

    def get_default_callables(self):
        """ Get a dictionary of default callables, in form {name:callable}. Re-defined in subclasses."""
        return {}

    #: Reference to System object
    system = Instance(SystemBase, transient=True)

    #: Description of the object (shown in WEB interface)
    description = CUnicode

    #: Python Logger instance for this object. System creates each object its own logger instance.
    logger = Instance(logging.Logger, transient=True)

    #: Tags are used for (for example) grouping objects. See :ref:`groups`.
    tags = TagSet(trait=CUnicode)

    #: Name property is determined by System namespace. Can be read/written.
    name = Property(trait=Unicode, depends_on='name_changed_event')

    @cached_property
    def _get_name(self):
        try:
            return self.system.reverse[self]
        except (KeyError, AttributeError):
            return 'System not initialized!'

    def _set_name(self, new_name):
        if not is_valid_variable_name(new_name):
            raise NameError('Illegal name %s' % new_name)
        try:
            if self in list(self.system.namespace.values()):
                del self.system.namespace[self.name]
        except NameError:

            pass
        self.system.namespace[new_name] = self
        self.logger = self.system.logger.getChild('%s.%s' % (self.__class__.__name__, new_name))

    #: If set to *True*, current SystemObject is hidden in the UML diagram of WEB interface.
    hide_in_uml = CBool(False)

    _order = Int
    _count = 0

    #: Attributes that can be edited by user in WEB interface
    view = ['hide_in_uml']

    #: The data type name (as string) of the object. This is written in the initialization, and is used by WEB
    #: interface Django templates.
    data_type = ''

    #: If editable=True, a quick edit widget will appear in the web interface. Define in subclasses.
    editable = False

    # Namespace triggers this event when object name name is changed
    name_changed_event = Event

    _passed_arguments = Tuple(transient=True)
    _postponed_callables = Dict(transient=True)

    @property
    def class_name(self):
        # For Django templates
        return self.__class__.__name__

    @property
    def object_type(self):
        """
            A read-only property that gives the object type as string; sensor, actuator, program, other.
            Used by WEB interface templates.
        """

        from .statusobject import AbstractSensor, AbstractActuator
        from .program import Program
        if isinstance(self, AbstractSensor):
            return 'sensor'
        elif isinstance(self, AbstractActuator):
            return 'actuator'
        elif isinstance(self, Program):
            return 'program'
        else:
            return 'other'

    def __init__(self, name='', **traits):
        # Postpone traits initialization to be launched by System
        self.logger = logging.getLogger('automate.%s' % self.__class__.__name__)
        self._order = SystemObject._count
        SystemObject._count += 1

        self._passed_arguments = name, traits
        if 'system' in traits:
            self.setup_system(traits.pop('system'))
            self.setup_callables()

    def __setstate__(self, state, trait_change_notify=True):
        self.logger = logging.getLogger('automate.%s' % self.__class__.__name__)
        self._order = state.pop('_order')
        self._passed_arguments = None, state

    def get_status_display(self, **kwargs):
        """
            Redefine this in subclasses if status can be represented in human-readable way (units etc.)
        """
        if 'value' in kwargs:
            return str(kwargs['value'])
        return self.class_name

    def get_as_datadict(self):
        """
            Get information about this object as a dictionary.  Used by WebSocket interface to pass some
            relevant information to client applications.
        """
        return dict(type=self.__class__.__name__, tags=list(self.tags))

    def setup(self, *args, **kwargs):
        """
            Initialize necessary services etc. here. Define this in subclasses.
        """
        pass

    def setup_system(self, system, name_from_system='', **kwargs):
        """
            Set system attribute and do some initialization. Used by System.
        """

        if not self.system:
            self.system = system
        name, traits = self._passed_arguments
        new_name = self.system.get_unique_name(self, name, name_from_system)
        if not self in self.system.reverse:
            self.name = new_name
        if name is None and 'name' in traits:  # Only __setstate__ sets name to None. Default is ''.
            del traits['name']

        for cname in self.callables:
            if cname in traits:
                c = self._postponed_callables[cname] = traits.pop(cname)
                c.setup_callable_system(self.system)
            getattr(self, cname).setup_callable_system(self.system)

        if not self.traits_inited():
            super(SystemObject, self).__init__(**traits)
        self.name_changed_event = True
        self.setup()

    def setup_callables(self):
        """
            Setup Callable attributes that belong to this object.
        """
        defaults = self.get_default_callables()
        for key, value in list(defaults.items()):
            self._postponed_callables.setdefault(key, value)
        for key in self.callables:
            value = self._postponed_callables.pop(key)
            value.setup_callable_system(self.system, init=True)
            setattr(self, key, value)

    def cleanup(self):
        """
            Write here whatever cleanup actions are needed when object is no longer used.
        """

    def __str__(self):
        return self.name

    def __repr__(self):
        return u"'%s'" % self.name
